import datetime
import os
import queue
import re
import threading
import time

from dateutil.tz import tzutc
from typing import Iterable, Optional, Collection, Set
from xml.etree import ElementTree

import dateutil.parser
import requests

from platypush.context import get_bus, get_plugin
from platypush.message.event.rss import NewFeedEntryEvent
from platypush.plugins import RunnablePlugin, action
from platypush.schemas.rss import RssFeedEntrySchema


class RssPlugin(RunnablePlugin):
    """
    A plugin for parsing and subscribing to RSS feeds.

    Triggers:

        - :class:`platypush.message.event.rss.NewFeedEntryEvent` when a new entry is received on a subscribed feed.

    Requires:

        * **feedparser** (``pip install feedparser``)
        * **defusedxml** (``pip install defusedxml``)

    """

    user_agent = (
        'Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) '
        + 'Chrome/62.0.3202.94 Safari/537.36'
    )

    def __init__(
        self,
        subscriptions: Optional[Collection[str]] = None,
        poll_seconds: int = 300,
        user_agent: str = user_agent,
        **kwargs,
    ):
        """
        :param subscriptions: List of feeds to monitor for updates, as URLs.
            OPML URLs/local files are also supported.
        :param poll_seconds: How often we should check for updates (default: 300 seconds).
        :param user_agent: Custom user agent to use for the requests.
        """
        super().__init__(**kwargs)
        self.poll_seconds = poll_seconds
        self.user_agent = user_agent
        self._feeds_metadata = {}
        self._feed_worker_queues = [queue.Queue()] * 5
        self._feed_response_queue = queue.Queue()
        self._feed_workers = []
        self._latest_entries = []

        self.subscriptions = list(self._parse_subscriptions(subscriptions or []))

        self._latest_timestamps = self._get_latest_timestamps()

    @staticmethod
    def _get_feed_latest_timestamp_varname(url: str) -> str:
        return f'LATEST_FEED_TIMESTAMP[{url}]'

    @classmethod
    def _get_feed_latest_timestamp(cls, url: str) -> Optional[datetime.datetime]:
        t = (
            get_plugin('variable')
            .get(cls._get_feed_latest_timestamp_varname(url))
            .output.get(cls._get_feed_latest_timestamp_varname(url))
        )

        if t:
            return dateutil.parser.isoparse(t)

    def _get_latest_timestamps(self) -> dict:
        return {url: self._get_feed_latest_timestamp(url) for url in self.subscriptions}

    def _update_latest_timestamps(self) -> None:
        variable = get_plugin('variable')
        variable.set(
            **{
                self._get_feed_latest_timestamp_varname(url): latest_timestamp
                for url, latest_timestamp in self._latest_timestamps.items()
            }
        )

    @staticmethod
    def _parse_content(entry) -> Optional[str]:
        content = getattr(entry, 'content', None)
        if not content:
            return

        if isinstance(content, list):
            return content[0]['value']

        return content

    @action
    def parse_feed(self, url: str):
        """
        Parse a feed URL.

        :param url: Feed URL.
        :return: .. schema:: rss.RssFeedEntrySchema(many=True)
        """
        import feedparser

        feed = feedparser.parse(
            requests.get(url, headers={'User-Agent': self.user_agent}).text
        )
        return RssFeedEntrySchema().dump(
            sorted(
                [
                    {
                        'feed_url': url,
                        'feed_title': getattr(feed.feed, 'title', None),
                        'id': getattr(entry, 'id', None),
                        'url': entry.link,
                        'published': datetime.datetime.fromtimestamp(
                            time.mktime(entry.published_parsed)
                        ),
                        'title': entry.title,
                        'summary': getattr(entry, 'summary', None),
                        'content': self._parse_content(entry),
                    }
                    for entry in feed.entries
                    if getattr(entry, 'published_parsed', None)
                ],
                key=lambda e: e['published'],
            ),
            many=True,
        )

    @action
    def get_latest_entries(self, limit: int = 20):
        """
        Get the latest entries from the subscribed feeds, sorted by descending published date.

        :param limit: Maximum number of entries to return (default: 20).
        :return: .. schema:: rss.RssFeedEntrySchema(many=True)
        """
        return sorted(self._latest_entries, key=lambda e: e['published'], reverse=True)[
            :limit
        ]

    def _feed_worker(self, q: queue.Queue):
        while not self.should_stop():
            try:
                url = q.get(block=True, timeout=1)
            except queue.Empty:
                continue

            try:
                self._feed_response_queue.put(
                    {
                        'url': url,
                        'content': self.parse_feed(url).output,
                    }
                )
            except Exception as e:
                self._feed_response_queue.put(
                    {
                        'url': url,
                        'error': e,
                    }
                )

        self._feed_response_queue.put(None)

    def _parse_opml_lists(self, subs: Iterable[str]) -> Set[str]:
        from defusedxml import ElementTree

        feeds = set()
        subs = set(subs)
        content_by_sub = {}
        urls = {sub for sub in subs if re.search(r'^https?://', sub)}
        files = {os.path.expanduser(sub) for sub in subs if sub not in urls}

        for url in urls:
            try:
                content_by_sub[url] = requests.get(
                    url,
                    headers={
                        'User-Agent': self.user_agent,
                    },
                ).text
            except Exception as e:
                self.logger.warning('Could not retrieve subscription %s: %s', url, e)

        for file in files:
            try:
                with open(file, 'r') as f:
                    content_by_sub[file] = f.read()
            except Exception as e:
                self.logger.warning('Could not open file %s: %s', file, e)

        for sub, content in content_by_sub.items():
            root = ElementTree.fromstring(content.strip())
            if root.tag != 'opml':
                self.logger.warning('%s is not a valid OPML resource', sub)
                continue

            feeds.update(self._parse_feeds_from_outlines(root.findall('body/outline')))

        return feeds

    def _parse_feeds_from_outlines(
        self,
        outlines: Iterable[ElementTree.Element],
    ) -> Set[str]:
        feeds = set()
        outlines = list(outlines)

        while outlines:
            outline = outlines.pop(0)
            if 'xmlUrl' in outline.attrib:
                url = outline.attrib['xmlUrl']
                feeds.add(url)
                self._feeds_metadata[url] = {
                    **self._feeds_metadata.get(url, {}),
                    'title': outline.attrib.get('title'),
                    'description': outline.attrib.get('text'),
                    'url': outline.attrib.get('htmlUrl'),
                }

            for i, child in enumerate(outline.iter()):
                if i > 0:
                    outlines.append(child)

        return feeds

    def _parse_subscriptions(self, subs: Iterable[str]) -> Iterable[str]:
        import feedparser

        self.logger.info('Parsing feed subscriptions')
        feeds = set()
        lists = set()

        for sub in subs:
            try:
                # Check if it's an OPML list of feeds or an individual feed
                feed = feedparser.parse(sub)
                if feed.feed.get('opml'):
                    lists.add(sub)
                else:
                    channel = feed.get('channel', {})
                    self._feeds_metadata[sub] = {
                        **self._feeds_metadata.get(sub, {}),
                        'title': channel.get('title'),
                        'description': channel.get('description'),
                        'url': channel.get('link'),
                    }

                    feeds.add(sub)
            except Exception as e:
                self.logger.warning('Could not parse %s: %s', sub, e)

        feeds.update(self._parse_opml_lists(lists))
        return feeds

    @staticmethod
    def _datetime_to_string(dt: datetime.datetime) -> str:
        return dt.replace(tzinfo=tzutc()).strftime('%a, %d %b %Y %H:%M:%S %Z')

    @action
    def export_to_opml(self) -> str:
        """
        Export the list of subscriptions into OPML format.

        :return: The list of subscriptions as a string in OPML format.
        """
        root = ElementTree.Element('opml', {'version': '2.0'})

        head = ElementTree.Element('head')
        title = ElementTree.Element('title')
        title.text = 'Platypush feed subscriptions'
        created = ElementTree.Element('dateCreated')
        created.text = self._datetime_to_string(datetime.datetime.utcnow())
        head.append(title)
        head.append(created)

        body = ElementTree.Element('body')
        feeds = ElementTree.Element('outline', {'text': 'Feeds'})

        for sub in self.subscriptions:
            metadata = self._feeds_metadata.get(sub, {})
            feed = ElementTree.Element(
                'outline',
                {
                    'xmlUrl': sub,
                    'text': metadata.get('description', metadata.get('title', sub)),
                    **({'htmlUrl': metadata['url']} if metadata.get('url') else {}),
                    **({'title': metadata['title']} if metadata.get('title') else {}),
                },
            )

            feeds.append(feed)

        body.append(feeds)

        root.append(head)
        root.append(body)
        return ElementTree.tostring(root, encoding='utf-8', method='xml').decode()

    def main(self):
        self._feed_workers = [
            threading.Thread(target=self._feed_worker, args=(q,))
            for q in self._feed_worker_queues
        ]

        for worker in self._feed_workers:
            worker.start()

        self.logger.info(
            f'Initialized RSS plugin with {len(self.subscriptions)} subscriptions'
        )

        while not self.should_stop():
            responses = {}
            for i, url in enumerate(self.subscriptions):
                worker_queue = self._feed_worker_queues[
                    i % len(self._feed_worker_queues)
                ]
                worker_queue.put(url)

            time_start = time.time()
            timeout = 60
            max_time = time_start + timeout
            new_entries = []

            while (
                not self.should_stop()
                and len(responses) < len(self.subscriptions)
                and time.time() - time_start <= timeout
            ):
                try:
                    response = self._feed_response_queue.get(
                        block=True, timeout=max_time - time_start
                    )
                except queue.Empty:
                    self.logger.warning('RSS parse timeout')
                    break

                if not response:
                    continue

                url = response['url']
                error = response.get('error')
                if error:
                    self.logger.error(f'Could not parse feed {url}: {error}')
                    responses[url] = error
                else:
                    responses[url] = response['content']

            responses = {
                k: v for k, v in responses.items() if not isinstance(v, Exception)
            }

            for url, response in responses.items():
                latest_timestamp = self._latest_timestamps.get(url)
                new_entries += response

                for entry in response:
                    published = datetime.datetime.fromisoformat(entry['published'])
                    if not latest_timestamp or published > latest_timestamp:
                        latest_timestamp = published
                        get_bus().post(NewFeedEntryEvent(**entry))

                self._latest_timestamps[url] = latest_timestamp

            self._update_latest_timestamps()
            self._latest_entries = new_entries
            self.wait_stop(self.poll_seconds)

    def stop(self):
        super().stop()
        for worker in self._feed_workers:
            worker.join(timeout=60)

        self.logger.info('RSS integration stopped')


# vim:sw=4:ts=4:et:



import datetime
import enum
import os
import time
import threading

from typing import Optional, Union, Dict, Any, List

import croniter
from dateutil.tz import gettz

from platypush.backend import Backend
from platypush.context import get_bus, get_plugin
from platypush.message.event.alarm import AlarmStartedEvent, AlarmDismissedEvent, AlarmSnoozedEvent
from platypush.plugins.media import MediaPlugin, PlayerState
from platypush.procedure import Procedure


class AlarmState(enum.IntEnum):
    WAITING = 1
    RUNNING = 2
    DISMISSED = 3
    SNOOZED = 4
    SHUTDOWN = 5


class Alarm:
    _alarms_count = 0
    _id_lock = threading.RLock()

    def __init__(self, when: str, actions: Optional[list] = None, name: Optional[str] = None,
                 audio_file: Optional[str] = None, audio_plugin: Optional[str] = None,
                 audio_volume: Optional[Union[int, float]] = None,
                 snooze_interval: float = 300.0, enabled: bool = True):
        with self._id_lock:
            self._alarms_count += 1
            self.id = self._alarms_count

        self.when = when
        self.name = name or 'Alarm_{}'.format(self.id)
        self.audio_file = None

        if audio_file:
            self.audio_file = os.path.abspath(os.path.expanduser(audio_file))
            assert os.path.isfile(self.audio_file), 'No such audio file: {}'.format(self.audio_file)

        self.audio_plugin = audio_plugin
        self.audio_volume = audio_volume
        self.snooze_interval = snooze_interval
        self.state: Optional[AlarmState] = None
        self.timer: Optional[threading.Timer] = None
        self.actions = Procedure.build(name=name, _async=False, requests=actions or [], id=self.id)

        self._enabled = enabled
        self._runtime_snooze_interval = snooze_interval

    def get_next(self) -> float:
        now = datetime.datetime.now().replace(tzinfo=gettz())  # lgtm [py/call-to-non-callable]

        try:
            cron = croniter.croniter(self.when, now)
            return cron.get_next()
        except (AttributeError, croniter.CroniterBadCronError):
            try:
                timestamp = datetime.datetime.fromisoformat(self.when).replace(
                    tzinfo=gettz())  # lgtm [py/call-to-non-callable]
            except (TypeError, ValueError):
                timestamp = (datetime.datetime.now().replace(tzinfo=gettz()) +  # lgtm [py/call-to-non-callable]
                             datetime.timedelta(seconds=int(self.when)))

            return timestamp.timestamp() if timestamp >= now else None

    def is_enabled(self):
        return self._enabled

    def disable(self):
        self._enabled = False

    def enable(self):
        self._enabled = True

    def dismiss(self):
        self.state = AlarmState.DISMISSED
        self.stop_audio()
        get_bus().post(AlarmDismissedEvent(name=self.name))

    def snooze(self, interval: Optional[float] = None):
        self._runtime_snooze_interval = interval or self.snooze_interval
        self.state = AlarmState.SNOOZED
        self.stop_audio()
        get_bus().post(AlarmSnoozedEvent(name=self.name, interval=self._runtime_snooze_interval))

    def start(self):
        if self.timer:
            self.timer.cancel()

        if self.get_next() is None:
            return

        interval = self.get_next() - time.time()
        self.timer = threading.Timer(interval, self.callback())
        self.timer.start()
        self.state = AlarmState.WAITING

    def stop(self):
        self.state = AlarmState.SHUTDOWN
        if self.timer:
            self.timer.cancel()
            self.timer = None

    def _get_audio_plugin(self) -> MediaPlugin:
        return get_plugin(self.audio_plugin)

    def play_audio(self):
        def thread():
            self._get_audio_plugin().play(self.audio_file)
            if self.audio_volume is not None:
                self._get_audio_plugin().set_volume(self.audio_volume)

        self.state = AlarmState.RUNNING
        audio_thread = threading.Thread(target=thread)
        audio_thread.start()

    def stop_audio(self):
        self._get_audio_plugin().stop()

    def callback(self):
        def _callback():
            while True:
                if self.state == AlarmState.SHUTDOWN:
                    break

                if self.is_enabled():
                    get_bus().post(AlarmStartedEvent(name=self.name))
                    if self.audio_plugin and self.audio_file:
                        self.play_audio()

                    self.actions.execute()

                time.sleep(10)
                sleep_time = None
                if self.state == AlarmState.RUNNING:
                    while True:
                        state = self._get_audio_plugin().status().output.get('state')
                        if state == PlayerState.STOP.value:
                            if self.state == AlarmState.SNOOZED:
                                sleep_time = self._runtime_snooze_interval
                            else:
                                self.state = AlarmState.WAITING

                            break
                        else:
                            time.sleep(10)

                if self.state == AlarmState.SNOOZED:
                    sleep_time = self._runtime_snooze_interval
                elif self.get_next() is None:
                    self.state = AlarmState.SHUTDOWN
                    break

                if not sleep_time:
                    sleep_time = self.get_next() - time.time() if self.get_next() else 10

                time.sleep(sleep_time)

        return _callback

    def to_dict(self):
        return {
            'name': self.name,
            'id': self.id,
            'when': self.when,
            'next_run': self.get_next(),
            'enabled': self.is_enabled(),
            'state': self.state.name,
        }


class AlarmBackend(Backend):
    """
    Backend to handle user-configured alarms.

    Triggers:

        * :class:`platypush.message.event.alarm.AlarmStartedEvent` when an alarm starts.
        * :class:`platypush.message.event.alarm.AlarmSnoozedEvent` when an alarm is snoozed.
        * :class:`platypush.message.event.alarm.AlarmTimeoutEvent` when an alarm times out.
        * :class:`platypush.message.event.alarm.AlarmDismissedEvent` when an alarm is dismissed.

    """

    def __init__(self, alarms: Optional[Union[list, Dict[str, Any]]] = None, audio_plugin: str = 'media.mplayer',
                 *args, **kwargs):
        """
        :param alarms: List or name->value dict with the configured alarms. Example:

        .. code-block:: yaml

            morning_alarm:
                when: '0 7 * * 1-5'   # Cron expression format: run every weekday at 7 AM
                audio_file: ~/path/your_ringtone.mp3
                audio_plugin: media.mplayer
                audio_volume: 10       # 10%
                snooze_interval: 300   # 5 minutes snooze
                actions:
                    - action: tts.say
                      args:
                          text: Good morning

                    - action: light.hue.bri
                      args:
                          value: 1

                    - action: light.hue.bri
                      args:
                          value: 140
                          transitiontime: 150

            one_shot_alarm:
                when: '2020-02-18T07:00:00.000000'   # One-shot execution, with timestamp in ISO format
                audio_file: ~/path/your_ringtone.mp3
                actions:
                    - action: light.hue.on

        :param audio_plugin: Media plugin (instance of :class:`platypush.plugins.media.MediaPlugin`) that will be
            used to play the alarm audio (default: ``media.mplayer``).
        """
        super().__init__(*args, **kwargs)
        alarms = alarms or []
        if isinstance(alarms, dict):
            alarms = [{'name': name, **alarm} for name, alarm in alarms.items()]

        self.audio_plugin = audio_plugin
        alarms = [Alarm(**{'audio_plugin': self.audio_plugin, **alarm}) for alarm in alarms]
        self.alarms: Dict[str, Alarm] = {alarm.name: alarm for alarm in alarms}

    def add_alarm(self, when: str, actions: list, name: Optional[str] = None, audio_file: Optional[str] = None,
                  audio_volume: Optional[Union[int, float]] = None, enabled: bool = True) -> Alarm:
        alarm = Alarm(when=when, actions=actions, name=name, enabled=enabled, audio_file=audio_file,
                      audio_plugin=self.audio_plugin, audio_volume=audio_volume)

        if alarm.name in self.alarms:
            self.logger.info('Overwriting existing alarm {}'.format(alarm.name))
            self.alarms[alarm.name].stop()

        self.alarms[alarm.name] = alarm
        self.alarms[alarm.name].start()
        return self.alarms[alarm.name]

    def _get_alarm(self, name) -> Alarm:
        assert name in self.alarms, 'Alarm {} does not exist'.format(name)
        return self.alarms[name]

    def enable_alarm(self, name: str):
        self._get_alarm(name).enable()

    def disable_alarm(self, name: str):
        self._get_alarm(name).disable()

    def dismiss_alarm(self):
        alarm = self.get_running_alarm()
        if not alarm:
            self.logger.info('No alarm is running')
            return

        alarm.dismiss()

    def snooze_alarm(self, interval: Optional[str] = None):
        alarm = self.get_running_alarm()
        if not alarm:
            self.logger.info('No alarm is running')
            return

        alarm.snooze(interval=interval)

    def get_alarms(self) -> List[Alarm]:
        return sorted([alarm for alarm in self.alarms.values()], key=lambda alarm: alarm.get_next())

    def get_running_alarm(self) -> Optional[Alarm]:
        running_alarms = [alarm for alarm in self.alarms.values() if alarm.state == AlarmState.RUNNING]
        return running_alarms[0] if running_alarms else None

    def __enter__(self):
        for alarm in self.alarms.values():
            alarm.stop()
            alarm.start()

        self.logger.info('Initialized alarm backend with {} alarms'.format(len(self.alarms)))

    def __exit__(self, exc_type, exc_val, exc_tb):
        for alarm in self.alarms.values():
            alarm.stop()

        self.logger.info('Alarm backend terminated')

    def loop(self):
        for name, alarm in self.alarms.copy().items():
            if not alarm.timer or (not alarm.timer.is_alive() and alarm.state == AlarmState.SHUTDOWN):
                del self.alarms[name]

        time.sleep(10)


# vim:sw=4:ts=4:et:



import json
from typing import Optional

from flask import Blueprint, request
from flask.wrappers import Response

from platypush.backend.http.app import template_folder
from platypush.backend.http.app.utils import authenticate
from platypush.context import get_plugin
from platypush.plugins.camera import CameraPlugin, Camera, StreamWriter

camera = Blueprint('camera', __name__, template_folder=template_folder)

# Declare routes list
__routes__ = [
    camera,
]


def get_camera(plugin: str) -> CameraPlugin:
    plugin_name = f'camera.{plugin}'
    p = get_plugin(plugin_name)
    assert p, f'No such plugin: {plugin_name}'
    return p


def get_frame(session: Camera, timeout: Optional[float] = None) -> Optional[bytes]:
    if session.stream:
        with session.stream.ready:
            session.stream.ready.wait(timeout=timeout)
            return session.stream.frame


def feed(camera: CameraPlugin, **kwargs):
    with camera.open(**kwargs) as session:
        camera.start_camera(session)
        while True:
            frame = get_frame(session, timeout=5.0)
            if frame:
                yield frame


def get_args(kwargs):
    kwargs = kwargs.copy()
    if 't' in kwargs:
        del kwargs['t']

    for k, v in kwargs.items():
        if k == 'resolution':
            v = json.loads('[{}]'.format(v))
        else:
            try:
                v = int(v)
            except (ValueError, TypeError):
                try:
                    v = float(v)
                except (ValueError, TypeError):
                    pass

        kwargs[k] = v

    return kwargs


@camera.route('/camera/<plugin>/photo.<extension>', methods=['GET'])
@authenticate()
def get_photo(plugin, extension):
    plugin = get_camera(plugin)
    extension = 'jpeg' if extension in ('jpg', 'jpeg') else extension

    with plugin.open(stream=True, stream_format=extension, frames_dir=None, **get_args(request.args)) as session:
        plugin.start_camera(session)
        frame = None
        for _ in range(session.info.warmup_frames):
            frame = get_frame(session)

        return Response(frame, mimetype=session.stream.mimetype)


@camera.route('/camera/<plugin>/video.<extension>', methods=['GET'])
@authenticate()
def get_video(plugin, extension):
    stream_class = StreamWriter.get_class_by_name(extension)
    camera = get_camera(plugin)
    return Response(
        feed(camera, stream=True, stream_format=extension, frames_dir=None,
            **get_args(request.args)
        ), mimetype=stream_class.mimetype
    )


@camera.route('/camera/<plugin>/photo', methods=['GET'])
@authenticate()
def get_photo_default(plugin):
    return get_photo(plugin, 'jpeg')


@camera.route('/camera/<plugin>/video', methods=['GET'])
@authenticate()
def get_video_default(plugin):
    return get_video(plugin, 'mjpeg')


@camera.route('/camera/<plugin>/frame', methods=['GET'])
@authenticate()
def get_photo_deprecated(plugin):
    return get_photo_default(plugin)


@camera.route('/camera/<plugin>/feed', methods=['GET'])
@authenticate()
def get_video_deprecated(plugin):
    return get_video_default(plugin)


# vim:sw=4:ts=4:et:



from platypush.message.response import Response


class SSHResponse(Response):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class SSHKeygenResponse(SSHResponse):
    def __init__(self, fingerprint: str, key_file: str, pub_key_file: str, *args, **kwargs):
        super().__init__(*args, output={
            'fingerprint': fingerprint,
            'key_file': key_file,
            'pub_key_file': pub_key_file,
        }, **kwargs)


# vim:sw=4:ts=4:et:



from typing import List, Optional

from platypush.plugins.lcd import LcdPlugin


class LcdGpioPlugin(LcdPlugin):
    """
    Plugin to write to an LCD display connected via GPIO.

    Requires:

        * **RPLCD** (``pip install RPLCD``)
        * **RPi.GPIO** (``pip install RPi.GPIO``)

    """

    def __init__(self, pin_rs: int, pin_e: int, pins_data: List[int],
                 pin_rw: Optional[int] = None, pin_mode: str = 'BOARD',
                 pin_backlight: Optional[int] = None,
                 cols: int = 16, rows: int = 2,
                 backlight_enabled: bool = True,
                 backlight_mode: str = 'active_low',
                 dotsize: int = 8, charmap: str = 'A02',
                 auto_linebreaks: bool = True,
                 compat_mode: bool = False, **kwargs):
        """
        :param pin_rs: Pin for register select (RS).
        :param pin_e: Pin to start data read or write (E).
        :param pins_data: List of data bus pins in 8 bit mode (DB0-DB7) or in 4
            bit mode (DB4-DB7) in ascending order.
        :param pin_mode: Which scheme to use for numbering of the GPIO pins,
            either ``BOARD`` or ``BCM``. Default: ``BOARD``.
        :param pin_rw: Pin for selecting read or write mode (R/W). Default:
            ``None``, read only mode.
        :param pin_backlight: Pin for controlling backlight on/off. Set this to
            ``None`` for no backlight control. Default: ``None``.
        :param cols: Number of columns per row (usually 16 or 20). Default: ``16``.
        :param rows: Number of display rows (usually 1, 2 or 4). Default: ``2``.
        :param backlight_enabled: Whether the backlight is enabled initially.
            Default: ``True``. Has no effect if pin_backlight is ``None``
        :param backlight_mode: Set this to either ``active_high`` or ``active_low``
            to configure the operating control for the backlight. Has no effect if
            pin_backlight is ``None``
        :param dotsize: Some 1 line displays allow a font height of 10px.
            Allowed: ``8`` or ``10``. Default: ``8``.
        :param charmap: The character map used. Depends on your LCD. This must
            be either ``A00`` or ``A02`` or ``ST0B``. Default: ``A02``.
        :param auto_linebreaks: Whether or not to automatically insert line
            breaks. Default: ``True``.
        :param compat_mode: Whether to run additional checks to support older LCDs
            that may not run at the reference clock (or keep up with it).
            Default: ``False``.
        """
        super().__init__(**kwargs)

        self.pin_mode = self._get_pin_mode(pin_mode)
        self.pin_rs = pin_rs
        self.pin_e = pin_e
        self.pin_rw = pin_rw
        self.pin_backlight = pin_backlight
        self.pins_data = pins_data
        self.cols = cols
        self.rows = rows
        self.backlight_enabled = backlight_enabled
        self.backlight_mode = backlight_mode
        self.dotsize = dotsize
        self.auto_linebreaks = auto_linebreaks
        self.compat_mode = compat_mode
        self.charmap = charmap

    def _get_lcd(self):
        from RPLCD.gpio import CharLCD
        return CharLCD(cols=self.cols, rows=self.rows, pin_rs=self.pin_rs,
                       pin_e=self.pin_e, pins_data=self.pins_data,
                       numbering_mode=self.pin_mode, pin_rw=self.pin_rw,
                       pin_backlight=self.pin_backlight,
                       backlight_enabled=self.backlight_enabled,
                       backlight_mode=self.backlight_mode,
                       dotsize=self.dotsize, charmap=self.charmap,
                       auto_linebreaks=self.auto_linebreaks,
                       compat_mode=self.compat_mode)


# vim:sw=4:ts=4:et:



import os
import re

from watchdog.events import FileSystemEventHandler, PatternMatchingEventHandler, RegexMatchingEventHandler

from platypush.backend.file.monitor.entities.resources import MonitoredResource, MonitoredPattern, MonitoredRegex
from platypush.context import get_bus
from platypush.message.event.file import FileSystemModifyEvent, FileSystemCreateEvent, FileSystemDeleteEvent


class EventHandler(FileSystemEventHandler):
    """
    Base class for Watchdog event handlers.
    """
    def __init__(self, resource: MonitoredResource, **kwargs):
        super().__init__(**kwargs)
        resource.path = os.path.expanduser(resource.path)
        self.resource = resource

    def _should_ignore_event(self, event) -> bool:
        ignore_dirs = [
            os.path.expanduser(
                _dir if os.path.expanduser(_dir).strip('/').startswith(self.resource.path)
                else os.path.join(self.resource.path, _dir)
            )
            for _dir in getattr(self.resource, 'ignore_directories', [])
        ]

        ignore_patterns = getattr(self.resource, 'ignore_patterns', None)
        ignore_regexes = getattr(self.resource, 'ignore_regexes', None)

        if ignore_dirs and any(
                event.src_path.startswith(ignore_dir) for ignore_dir in ignore_dirs
        ):
            return True

        if ignore_patterns and any(
                re.match(r'^{}$'.format(pattern.replace('*', '.*')), event.src_path)
                for pattern in ignore_patterns
        ):
            return True

        if ignore_regexes and any(
                re.match(regex, event.src_path)
                for regex in ignore_patterns
        ):
            return True

        return False

    def _on_event(self, event, output_event_type):
        if self._should_ignore_event(event):
            return
        get_bus().post(output_event_type(path=event.src_path, is_directory=event.is_directory))

    def on_created(self, event):
        self._on_event(event, FileSystemCreateEvent)

    def on_deleted(self, event):
        self._on_event(event, FileSystemDeleteEvent)

    def on_modified(self, event):
        self._on_event(event, FileSystemModifyEvent)

    def on_moved(self, event):
        pass

    @classmethod
    def from_resource(cls, resource: MonitoredResource):
        if isinstance(resource, MonitoredPattern):
            return PatternEventHandler(resource)
        if isinstance(resource, MonitoredRegex):
            return RegexEventHandler(resource)
        return cls(resource)


class PatternEventHandler(EventHandler, PatternMatchingEventHandler):
    """
    Event handler for file patterns.
    """
    def __init__(self, resource: MonitoredPattern):
        super().__init__(resource=resource,
                         patterns=resource.patterns,
                         ignore_patterns=resource.ignore_patterns,
                         ignore_directories=resource.ignore_directories,
                         case_sensitive=resource.case_sensitive)


class RegexEventHandler(EventHandler, RegexMatchingEventHandler):
    """
    Event handler for regex-based file patterns.
    """
    def __init__(self, resource: MonitoredRegex):
        super().__init__(resource=resource,
                         regexes=resource.regexes,
                         ignore_regexes=resource.ignore_regexes,
                         ignore_directories=resource.ignore_directories,
                         case_sensitive=resource.case_sensitive)



from typing import Optional

from platypush.message.event import Event
from platypush.plugins.mail import Mail


class MailEvent(Event):
    def __init__(self, mailbox: str, message: Optional[Mail] = None, *args, **kwargs):
        super().__init__(*args, mailbox=mailbox, message=message or {}, **kwargs)


class MailReceivedEvent(MailEvent):
    """
    Triggered when a new email is received.
    """
    pass


class MailSeenEvent(MailEvent):
    """
    Triggered when a previously unseen email is seen.
    """
    pass


class MailFlaggedEvent(MailEvent):
    """
    Triggered when a message is marked as flagged/starred.
    """
    pass


class MailUnflaggedEvent(MailEvent):
    """
    Triggered when a message previously marked as flagged/starred is unflagged.
    """
    pass


# vim:sw=4:ts=4:et:



import datetime
import json
import os
import re
import subprocess
import tempfile
from urllib.parse import urlparse

from platypush.plugins import action
from platypush.plugins.http.request import Plugin


class HttpWebpagePlugin(Plugin):
    """
    Plugin to handle and parse/simplify web pages.
    It used to use the Mercury Reader web API, but now that the API is discontinued this plugin is basically a
    wrapper around the `mercury-parser <https://github.com/postlight/mercury-parser>`_ JavaScript library.

    Requires:

        * **weasyprint** (``pip install weasyprint``), optional, for HTML->PDF conversion
        * **node** and **npm** installed on your system (to use the mercury-parser interface)
        * The mercury-parser library installed (``npm install -g @postlight/mercury-parser``)

    """

    _mercury_script = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'mercury-parser.js')

    @staticmethod
    def _parse(proc):
        with subprocess.Popen(proc, stdout=subprocess.PIPE, stderr=None) as parser:
            return parser.communicate()[0].decode()

    @staticmethod
    def _fix_relative_links(markdown: str, url: str) -> str:
        url = urlparse(url)
        base_url = f'{url.scheme}://{url.netloc}'
        return re.sub(r'(\[.+?])\((/.+?)\)', fr'\1({base_url}\2)', markdown)

    # noinspection PyShadowingBuiltins
    @action
    def simplify(self, url, type='html', html=None, outfile=None):
        """
        Parse the readable content of a web page removing any extra HTML elements using Mercury.

        :param url: URL to parse.
        :param type: Output format. Supported types: ``html``, ``markdown``, ``text`` (default: ``html``).
        :param html: Set this parameter if you want to parse some HTML content already fetched. Note
            that URL is still required by Mercury to properly style the output, but it won't be used
            to actually fetch the content.

        :param outfile: If set then the output will be written to the specified file. If the file extension
            is ``.pdf`` then the content will be exported in PDF format. If the output ``type`` is not
            specified then it can also be inferred from the extension of the output file.
        :return: dict

        Example if outfile is not specified::

            {
                "url": <url>,
                "title": <page title>,
                "content": <page parsed content>

            }

        Example if outfile is specified::

            {
                "url": <url>,
                "title": <page title>,
                "outfile": <output file absolute path>

            }

        """

        self.logger.info('Parsing URL {}'.format(url))
        wants_pdf = False

        if outfile:
            wants_pdf = outfile.lower().endswith('.pdf')
            if (
                    wants_pdf   # HTML will be exported to PDF
                    or outfile.lower().split('.')[-1].startswith('htm')
            ):
                type = 'html'
            elif outfile.lower().endswith('.md'):
                type = 'markdown'
            elif outfile.lower().endswith('.txt'):
                type = 'text'

        proc = ['node', self._mercury_script, url, type]
        f = None

        if html:
            f = tempfile.NamedTemporaryFile('w+', delete=False)
            f.write(html)
            f.flush()
            proc.append(f.name)

        try:
            response = self._parse(proc)
        finally:
            if f:
                os.unlink(f.name)

        try:
            response = json.loads(response.strip())
        except Exception as e:
            raise RuntimeError('Could not parse JSON: {}. Response: {}'.format(str(e), response))

        if type == 'markdown':
            response['content'] = self._fix_relative_links(response['content'], url)

        self.logger.debug('Got response from Mercury API: {}'.format(response))
        title = response.get('title', '{} on {}'.format(
            'Published' if response.get('date_published') else 'Generated',
            response.get('date_published', datetime.datetime.now().isoformat())))

        content = response.get('content', '')

        if not outfile:
            return {
                'url': url,
                'title': title,
                'content': content,
            }

        outfile = os.path.abspath(os.path.expanduser(outfile))
        style = '''
            body {
                font-size: 22px;
                font-family: 'Merriweather', Georgia, 'Times New Roman', Times, serif;
            }
            '''

        if type == 'html':
            content = (
                '''
                    <h1><a href="{url}" target="_blank">{title}</a></h1>
                    <div class="_parsed-content">{content}</div>
                '''.format(title=title, url=url, content=content)
            )

            if not wants_pdf:
                content = '''<html>
                        <head>
                            <title>{title}</title>
                            <style>{style}</style>
                        </head>'''.format(title=title, style=style) + \
                          '<body>{{' + content + '}}</body></html>'
        elif type == 'markdown':
            content = '# [{title}]({url})\n\n{content}'.format(
                title=title, url=url, content=content
            )

        if wants_pdf:
            import weasyprint
            try:
                from weasyprint.fonts import FontConfiguration
            except ImportError:
                from weasyprint.document import FontConfiguration

            font_config = FontConfiguration()
            css = [weasyprint.CSS('https://fonts.googleapis.com/css?family=Merriweather'),
                   weasyprint.CSS(string=style, font_config=font_config)]

            weasyprint.HTML(string=content).write_pdf(outfile, stylesheets=css)
        else:
            with open(outfile, 'w', encoding='utf-8') as f:
                f.write(content)

        return {
            'url': url,
            'title': title,
            'outfile': outfile,
        }


# vim:sw=4:ts=4:et:



import json
import queue

from datetime import datetime
from threading import Timer
from typing import Optional, List, Any, Dict, Union, Iterable, Callable

from platypush.message.event.zwave import ZwaveNodeRenamedEvent, ZwaveNodeEvent

from platypush.context import get_backend, get_bus
from platypush.message.response import Response
from platypush.plugins.mqtt import MqttPlugin, action
from platypush.plugins.zwave._base import ZwaveBasePlugin
from platypush.plugins.zwave._constants import command_class_by_name

_NOT_IMPLEMENTED_ERR = NotImplementedError('Not implemented by zwave.mqtt')


class ZwaveMqttPlugin(MqttPlugin, ZwaveBasePlugin):
    """
    This plugin allows you to manage a Z-Wave network over MQTT through
    `zwavejs2mqtt <https://github.com/zwave-js/zwavejs2mqtt>`_.

    Configuration required on the zwavejs2mqtt gateway:

        * Install the gateway following the instructions reported
          `here <https://zwave-js.github.io/zwavejs2mqtt/#/getting-started/quick-start>`_.

        * Configure the following settings from the zwavejs2mqtt web panel:

            * Zwave -> Serial Port: The path to your Z-Wave adapter.
            * Disabled MQTT Gateway: Set to false.
            * Configure the MQTT server that should be used by the service.
            * Mqtt -> Name: Unique name that identifies your gateway.
            * Mqtt -> Prefix: Prefix name for the topics published to MQTT.
            * Gateway -> Topic type: Set to "ValueID topics".
            * Gateway -> Payload type: Set to "JSON Time-Value".
            * Gateway -> Use nodes name instead of numeric nodeIDs: Set to false.
            * Gateway -> Send Zwave Events: Set to true.
            * Gateway -> Include Node Info: Set to true.

    Requires:

        * **paho-mqtt** (``pip install paho-mqtt``)

    """

    def __init__(self, name: str, host: str = 'localhost', port: int = 1883, topic_prefix: str = 'zwave',
                 timeout: int = 10, tls_certfile: Optional[str] = None, tls_keyfile: Optional[str] = None,
                 tls_version: Optional[str] = None, tls_ciphers: Optional[str] = None, username: Optional[str] = None,
                 password: Optional[str] = None, **kwargs):
        """
        :param name: Gateway name, as configured from the zwavejs2mqtt web panel from Mqtt -> Name.
        :param host: MQTT broker host, as configured from the zwavejs2mqtt web panel from Mqtt -> Host
            (default: ``localhost``).
        :param port: MQTT broker listen port, as configured from the zwavejs2mqtt web panel from Mqtt -> Port
            (default: 1883).
        :param topic_prefix: MQTT topic prefix, as specified from the zwavejs2mqtt web panel from Mqtt -> Prefix
            (default: ``zwave``).
        :param timeout: If the command expects from a response, then this timeout value will be used
            (default: 60 seconds).
        :param tls_cafile: If the connection requires TLS/SSL, specify the certificate authority file
            (default: None)
        :param tls_certfile: If the connection requires TLS/SSL, specify the certificate file (default: None)
        :param tls_keyfile: If the connection requires TLS/SSL, specify the key file (default: None)
        :param tls_version: If the connection requires TLS/SSL, specify the minimum TLS supported version
            (default: None)
        :param tls_ciphers: If the connection requires TLS/SSL, specify the supported ciphers (default: None)
        :param username: If the connection requires user authentication, specify the username (default: None)
        :param password: If the connection requires user authentication, specify the password (default: None)
        """

        super().__init__(host=host, port=port, tls_certfile=tls_certfile, tls_keyfile=tls_keyfile,
                         tls_version=tls_version, tls_ciphers=tls_ciphers, username=username,
                         password=password, **kwargs)

        self.topic_prefix = topic_prefix
        self.base_topic = topic_prefix + '/{}/ZWAVE_GATEWAY-' + name
        self.events_topic = self.base_topic.format('_EVENTS')
        self.timeout = timeout
        self._info = {
            'devices': {},
            'groups': {},
        }

        self._nodes_cache = {
            'by_id': {},
            'by_name': {},
        }

        self._values_cache = {
            'by_id': {},
            'by_label': {},
        }

        self._scenes_cache = {
            'by_id': {},
            'by_label': {},
        }

        self._groups_cache = {}

    @staticmethod
    def _get_backend():
        backend = get_backend('zwave.mqtt')
        if not backend:
            raise AssertionError('zwave.mqtt backend not configured')
        return backend

    def _api_topic(self, api: str) -> str:
        return self.base_topic.format('_CLIENTS') + '/api/{}'.format(api)

    def _topic(self, topic):
        return self.base_topic + '/' + topic

    @staticmethod
    def _parse_response(response: Union[dict, Response]) -> dict:
        if isinstance(response, Response):
            response = response.output

        assert response.get('success') is True, response.get('message', 'zwavejs2mqtt error')
        return response

    def _api_request(self, api: str, *args, **kwargs):
        payload = json.dumps({'args': args})
        ret = self._parse_response(
            self.publish(topic=self._api_topic(api) + '/set', msg=payload, reply_topic=self._api_topic(api),
                         **self._mqtt_args(**kwargs)))

        assert not ret or ret.get('success') is True, ret.get('message')
        return ret.get('result')

    @staticmethod
    def _convert_timestamp(t: Optional[int]) -> Optional[datetime]:
        if t:
            return datetime.fromtimestamp(t / 1000)

    def _get_scene(self, scene_id: Optional[int] = None, scene_label: Optional[str] = None, **kwargs) -> dict:
        assert scene_id or scene_label, 'No scene_id/scene_label specified'
        if scene_id in self._scenes_cache['by_id']:
            return self._scenes_cache['by_id'][scene_id]
        if scene_label in self._scenes_cache['by_label']:
            return self._scenes_cache['by_label'][scene_label]

        # noinspection PyUnresolvedReferences
        scenes = self.get_scenes(**kwargs).output
        scene = None

        if scene_id in scenes:
            scene = scenes[scene_id]
            self._scenes_cache['by_id'][scene_id] = scene
        else:
            scenes = [s for s in scenes if s['label'] == scene_label]
            if scenes:
                scene = scenes[0]
                self._scenes_cache['by_label'][scene_label] = scene

        assert scene, f'No such scene: scene_id={scene_id}, scene_label={scene_label}'
        return scene

    @classmethod
    def group_to_dict(cls, group: dict, node_id: Optional[int] = None) -> dict:
        return {
            'label': group['text'],
            'index': group['value'],
            'node_id': node_id,
            'group_id': f'{node_id}:{group["value"]}' if node_id is not None else None,
            'multichannel': group['multiChannel'],
            'max_associations': group['maxNodes'],
        }

    @classmethod
    def _get_type(cls, value: dict) -> str:
        if value['list']:
            if len(value.get('states', [])) > 1:
                return 'List'
            if value.get('min') is not None and value.get('max') is not None:
                return 'Decimal'
        if value['type'] == 'boolean':
            return 'Bool'
        if value['type'] == 'number':
            return 'Decimal'
        if value['type'] == 'string':
            return 'String'
        return value['type'].capitalize()

    @classmethod
    def value_to_dict(cls, value: dict) -> dict:
        return {
            'id': value['id'],
            'id_on_network': value['id'],
            'value_id': value['id'],
            'data': value.get('value'),
            'data_items': [state for state in value['states']] if len(value.get('states', [])) > 1 else None,
            'label': value.get('label', value.get('propertyName', value.get('property'))),
            'property_id': value.get('property'),
            'help': value.get('description'),
            'node_id': value.get('nodeId'),
            'parent_id': value.get('nodeId'),
            'type': cls._get_type(value),
            'default': value.get('default'),
            'endpoint': value.get('endpoint'),
            'stateless': value.get('stateless'),
            'command_class': value.get('commandClass'),
            'command_class_name': value.get('commandClassName'),
            'units': value.get('unit'),
            'min': value.get('min'),
            'max': value.get('max'),
            'is_read_only': value['readable'] and not value['writeable'],
            'is_write_only': value['writeable'] and not value['readable'],
            'last_update': cls._convert_timestamp(value.get('lastUpdate')),
            **({'property_key': value['propertyKey']} if 'propertyKey' in value else {}),
            **({'property_key_name': value['propertyKeyName']} if 'propertyKeyName' in value else {}),
        }

    @staticmethod
    def _get_value_id(value: dict) -> str:
        return f'{value["nodeId"]}-{value["commandClass"]}-{value.get("endpoint", 0)}-{value["property"]}'

    @classmethod
    def scene_to_dict(cls, scene: dict) -> dict:
        return {
            'scene_id': scene.pop('sceneid', None),
            'values': {
                cls._get_value_id(value): {
                    'node_id': value['nodeId'],
                    'value_id': cls._get_value_id(value),
                    'id_on_network': cls._get_value_id(value),
                    'data': value.get('value'),
                }
                for value in scene.pop('values', [])
            },
            **scene,
        }

    @classmethod
    def node_to_dict(cls, node: dict) -> dict:
        capabilities = []
        if node['supportsBeaming']:
            capabilities += ['beaming']
        if node['supportsSecurity']:
            capabilities += ['security']
        if node['isRouting']:
            capabilities += ['routing']
        if node.get('zwavePlusVersion'):
            capabilities += ['zwave_plus']

        return {
            'node_id': node['id'],
            'device_id': node['hexId'].replace('0x', ''),
            'name': node.get('name'),
            'capabilities': capabilities,
            'manufacturer_id': '0x{:04x}'.format(node['manufacturerId']) if node.get('manufacturerId') else None,
            'manufacturer_name': node.get('manufacturer'),
            'location': node.get('loc'),
            'status': node.get('status'),
            'is_available': node.get('available'),
            'is_awake': node.get('ready') and node.get('status') != 'Asleep',
            'is_beaming_device': node.get('supportsBeaming'),
            'is_controller': node.get('isControllerNode'),
            'is_failed': node.get('failed'),
            'is_frequent_listening_device': node.get('isFrequentListening'),
            'is_info_received': node.get('interviewStage') == 'Complete',
            'is_initialized': node.get('inited'),
            'is_listening_device': node.get('isListening'),
            'is_ready': node.get('ready'),
            'is_routing_device': node.get('isRouting'),
            'is_security_device': node.get('supportsSecurity'),
            'is_sleeping': node.get('ready') and node.get('status') == 'Asleep',
            'last_update': cls._convert_timestamp(node.get('lastActive')),
            'product_id': '0x{:04x}'.format(node['productId']) if node.get('productId') else None,
            'product_type': '0x{:04x}'.format(node['productType']) if node.get('productType') else None,
            'product_name': ' '.join([node.get('productLabel', ''), node.get('productDescription', '')]),
            'baud_rate': node.get('dataRate'),
            'max_baud_rate': node.get('maxBaudRate', node.get('dataRate')),
            'device_class': node.get('deviceClass'),
            'specific': node.get('deviceClass', {}).get('specific'),
            'firmware_version': node.get('firmwareVersion'),
            'keep_awake': node.get('keepAwake'),
            'db_link': node.get('dbLink'),
            'type': node.get('zwavePlusNodeType'),
            'role': node.get('zwavePlusRoleType'),
            'zwave_plus_version': node.get('zwavePlusVersion'),
            'zwave_plus_node_type': node.get('zwavePlusNodeType'),
            'zwave_plus_role_type': node.get('zwavePlusRoleType'),
            'neighbours': node.get('neighbors', []),

            'command_classes': {
                value['commandClassName']
                for value in node.get('values', {}).values() if value.get('commandClassName')
            },

            'groups': {
                group['value']: cls.group_to_dict(group, node_id=node['id'])
                for group in node.get('groups', [])
            },

            'values': {
                value['id']: cls.value_to_dict(value)
                for value in node.get('values', {}).values()
            },
        }

    def _get_node(self, node_id: Optional[int] = None, node_name: Optional[str] = None, use_cache: bool = True,
                  **kwargs) -> Optional[dict]:
        assert node_id or node_name, 'Please provide either a node_id or node_name'
        if use_cache:
            if node_id and node_id in self._nodes_cache['by_id']:
                return self._nodes_cache['by_id'][node_id]
            if node_name and node_name in self._nodes_cache['by_name']:
                return self._nodes_cache['by_name'][node_name]

        response = {
            node['id']: self.node_to_dict(node)
            for node in self._api_request('getNodes', **kwargs)
        }

        node = None
        if node_id:
            node = response.get(node_id)
        else:
            ret = [node for node in response.values() if node['name'] == node_name]
            if ret:
                node = ret[0]

        if node:
            self._nodes_cache['by_id'][node['node_id']] = node
            if node['name']:
                self._nodes_cache['by_name'][node['name']] = node

            for value in node.get('values', {}).values():
                self._values_cache['by_id'][value['id']] = value
                if value['label']:
                    self._values_cache['by_label'][value['label']] = value

        return node

    def _get_value(self, value_id: Optional[int] = None, id_on_network: Optional[str] = None,
                   value_label: Optional[str] = None, node_id: Optional[int] = None, node_name: Optional[str] = None,
                   use_cache: bool = True, **_) -> Dict[str, Any]:
        # Unlike python-openzwave, value_id and id_on_network are the same on zwavejs2mqtt
        value_id = value_id or id_on_network
        assert value_id or value_label, 'Please provide either value_id, id_on_network or value_label'

        if use_cache:
            if value_id and value_id in self._values_cache['by_id']:
                return self._values_cache['by_id'][value_id]
            if value_label and value_label in self._values_cache['by_label']:
                return self._values_cache['by_label'][value_label]

        nodes = []
        if node_id or node_name:
            nodes = [self._get_node(node_id=node_id, node_name=node_name, use_cache=False)]

        if not nodes:
            # noinspection PyUnresolvedReferences
            nodes = self.get_nodes().output
            assert nodes, 'No nodes found on the network'
            nodes = nodes.values()

        if value_id:
            values = [
                node['values'][value_id]
                for node in nodes
                if value_id in node.get('values', {})
            ]
        else:
            values = [
                value
                for node in nodes
                for value in node.get('values', {}).values()
                if value.get('label') == value_label
            ]

        assert values, f'No such value: {value_id or value_label}'
        value = values[0]

        if value.get('property_id') == 'targetValue':
            cur_value_id = '-'.join(value['value_id'].split('-')[:-1] + ['currentValue'])
            cur_value = self._nodes_cache['by_id'][value['node_id']].get('values', {}).get(cur_value_id)
            if cur_value:
                value['data'] = cur_value['data']

        self._values_cache['by_id'][value['id']] = value
        if value['label']:
            self._values_cache['by_label'][value['label']] = value

        return value

    def _topic_by_value_id(self, value_id: str) -> str:
        return self.topic_prefix + '/' + '/'.join(value_id.split('-'))

    def _filter_values(self, command_classes: Iterable[str], filter_callback: Optional[Callable[[dict], bool]] = None,
                       node_id: Optional[int] = None, node_name: Optional[str] = None, **kwargs) -> Dict[str, Any]:
        # noinspection PyUnresolvedReferences
        nodes = [self._get_node(node_name=node_name, use_cache=False, **kwargs)] if node_id or node_name else \
            self.get_nodes(**kwargs).output.values()

        command_classes = {
            command_class_by_name[command_name]
            for command_name in command_classes
        }

        values = {}

        for node in nodes:
            for value in node.get('values', {}).values():
                if value.get('command_class') not in command_classes or (
                        filter_callback and not filter_callback(value)):
                    continue

                value_id = value['id_on_network']
                if value_id.split('-').pop() == 'targetValue':
                    value_id = '-'.join(value_id.split('-')[:-1]) + '-currentValue'
                    cur_value = self._nodes_cache['by_id'][value['node_id']].get('values', {}).get(value_id)
                    if cur_value:
                        value['data'] = cur_value['data']

                values[value['id_on_network']] = value

        return values

    def _get_group(self, group_id: Optional[str] = None, group_index: Optional[int] = None, **kwargs) -> dict:
        group_id = group_id or group_index
        assert group_id is not None, 'No group_id/group_index specified'
        group = self._groups_cache.get(group_id)
        if group:
            return group

        # noinspection PyUnresolvedReferences
        groups = self.get_groups(**kwargs).output
        assert group_id in groups, f'No such group_id: {group_id}'
        return groups[group_id]

    @action
    def start_network(self, **kwargs):
        """
        Start the network (not implemented by zwavejs2mqtt).
        """
        raise _NOT_IMPLEMENTED_ERR

    @action
    def stop_network(self, **kwargs):
        """
        Stop the network (not implemented by zwavejs2mqtt).
        """
        raise _NOT_IMPLEMENTED_ERR

    @action
    def status(self, **kwargs) -> Dict[str, Any]:
        """
        Get the status of the controller.

        :param kwargs: Extra arguments to be passed to :meth:`platypush.plugins.mqtt.MqttPlugin.publish``
            (default: query the default configured device).
        :return: dict with the following fields: ``device`` and ``state``.
        """
        msg_queue = queue.Queue()
        topic = f'{self.topic_prefix}/Controller/status'
        client = self._get_client(**kwargs)

        def on_message(_, __, msg):
            if msg.topic != topic:
                return
            msg_queue.put(json.loads(msg.payload))

        client.on_message = on_message
        client.connect(kwargs.get('host', self.host), kwargs.get('port', self.port),
                       keepalive=kwargs.get('timeout', self.timeout))

        client.subscribe(topic)
        client.loop_start()

        try:
            status = msg_queue.get(block=True, timeout=kwargs.get('timeout', self.timeout))
        except queue.Empty:
            raise TimeoutError('The request timed out')
        finally:
            client.loop_stop()

        return {
            'device': status['nodeId'],
            'state': status['status'],
            'stats': {},
        }

    @action
    def add_node(self, do_security: bool = False, timeout: int = 60, **kwargs):
        """
        Start the inclusion process to add a node to the network.

        :param do_security: Whether to initialize the Network Key on the device if it supports the Security CC
        :param timeout: How long the inclusion process should last, in seconds (default: 60). Specify zero or null
            for no timeout.
        :param kwargs: Extra arguments to be passed to :meth:`platypush.plugins.mqtt.MqttPlugin.publish``
            (default: query the default configured device).
        """
        self._api_request('startInclusion', **kwargs)
        if timeout:
            Timer(timeout, lambda: self._api_request('stopInclusion', **kwargs)).start()

    @action
    def remove_node(self, **kwargs):
        """
        Remove a node from the network (not implemented on zwavejs2mqtt).

        :param kwargs: Extra arguments to be passed to :meth:`platypush.plugins.mqtt.MqttPlugin.publish``
            (default: query the default configured device).
        """
        raise _NOT_IMPLEMENTED_ERR

    @action
    def remove_failed_node(self, node_id: Optional[int] = None, node_name: Optional[str] = None, **kwargs):
        """
        Remove a failed node from the network.

        :param node_id: Filter by node_id.
        :param node_name: Filter by node name.
        :param kwargs: Extra arguments to be passed to :meth:`platypush.plugins.mqtt.MqttPlugin.publish``
            (default: query the default configured device).
        """
        if node_name:
            node_id = (self._get_node(node_name=node_name) or {}).get('node_id')

        assert node_id is not None, f'No such node_id: {node_id}'
        self._api_request('removeFailedNode', node_id, **kwargs)

    @action
    def replace_failed_node(self, node_id: Optional[int] = None, node_name: Optional[str] = None, **kwargs):
        """
        Replace a failed node on the network.

        :param node_id: Filter by node_id.
        :param node_name: Filter by node name.
        :param kwargs: Extra arguments to be passed to :meth:`platypush.plugins.mqtt.MqttPlugin.publish``
            (default: query the default configured device).
        """
        if node_name:
            node_id = (self._get_node(node_name=node_name) or {}).get('node_id')

        assert node_id is not None, f'No such node_id: {node_id}'
        self._api_request('replaceFailedNode', node_id, **kwargs)

    @action
    def replication_send(self, **kwargs):
        """
        Send node information from the primary to the secondary controller (not implemented by zwavejs2mqtt).
        """
        raise _NOT_IMPLEMENTED_ERR

    @action
    def request_network_update(self, node_id: Optional[int] = None, node_name: Optional[str] = None, **kwargs):
        """
        Request a network update to a node.

        :param node_id: Filter by node_id.
        :param node_name: Filter by node name.
        :param kwargs: Extra arguments to be passed to :meth:`platypush.plugins.mqtt.MqttPlugin.publish``
            (default: query the default configured device).
        """
        if node_name:
            node_id = (self._get_node(node_name=node_name) or {}).get('node_id')

        assert node_id is not None, f'No such node_id: {node_id}'
        self._api_request('refreshInfo', node_id, **kwargs)

    @action
    def request_node_neighbour_update(self, **kwargs):
        """
        Request a neighbours list update.

        :param kwargs: Extra arguments to be passed to :meth:`platypush.plugins.mqtt.MqttPlugin.publish``
            (default: query the default configured device).
        """
        self._api_request('refreshNeighbors', **kwargs)

    @action
    def get_nodes(self, node_id: Optional[int] = None, node_name: Optional[str] = None, **kwargs) \
            -> Optional[Dict[str, Any]]:
        """
        Get the nodes associated to the network.

        :param node_id: Filter by node_id.
        :param node_name: Filter by node name.
        :param kwargs: Extra arguments to be passed to :meth:`platypush.plugins.mqtt.MqttPlugin.publish``
            (default: query the default configured device).

        :return: List of paired devices. Example output:

            .. code-block:: json

                {
                  "1": {
                    "node_id": 1,
                    "device_id": "0115-0400-0001",
                    "name": "Controller",
                    "capabilities": [
                      "beaming",
                      "routing"
                    ],
                    "manufacturer_id": "0x0115",
                    "manufacturer_name": "Z-Wave.Me",
                    "location": "Living Room",
                    "status": "Alive",
                    "is_available": true,
                    "is_awake": true,
                    "is_beaming_device": true,
                    "is_controller": true,
                    "is_failed": false,
                    "is_frequent_listening_device": false,
                    "is_info_received": true,
                    "is_initialized": true,
                    "is_listening_device": true,
                    "is_ready": true,
                    "is_routing_device": true,
                    "is_security_device": false,
                    "is_sleeping": false,
                    "last_update": "2021-04-05T19:38:07.202000",
                    "product_id": null,
                    "product_type": null,
                    "product_name": "UZB Z-Wave USB Stick",
                    "baud_rate": 100000,
                    "max_baud_rate": null,
                    "device_class": {
                      "basic": 2,
                      "generic": 2,
                      "specific": 1
                    },
                    "specific": 1,
                    "firmware_version": null,
                    "keep_awake": false,
                    "db_link": "https://devices.zwave-js.io/?jumpTo=0x0115:0x0400:0x0001:0.0",
                    "zwave_plus_version": null,
                    "zwave_plus_node_type": null,
                    "zwave_plus_role_type": null,
                    "neighbours": [
                      3,
                      4
                    ],
                    "groups": {},
                    "values": {}
                  },
                  "4": {
                    "node_id": 4,
                    "device_id": "010f-0602-1003",
                    "name": "TV Switch",
                    "capabilities": [
                      "beaming",
                      "routing",
                      "zwave_plus"
                    ],
                    "manufacturer_id": "0x010f",
                    "manufacturer_name": "Fibargroup",
                    "location": "Living Room",
                    "status": "Alive",
                    "is_available": true,
                    "is_awake": true,
                    "is_beaming_device": true,
                    "is_controller": false,
                    "is_failed": false,
                    "is_frequent_listening_device": false,
                    "is_info_received": true,
                    "is_initialized": true,
                    "is_listening_device": true,
                    "is_ready": true,
                    "is_routing_device": true,
                    "is_security_device": false,
                    "is_sleeping": false,
                    "last_update": "2021-04-06T00:07:44.081000",
                    "product_id": null,
                    "product_type": null,
                    "product_name": "Metered Wall Plug Switch",
                    "baud_rate": 100000,
                    "max_baud_rate": null,
                    "device_class": {
                      "basic": 4,
                      "generic": 16,
                      "specific": 1
                    },
                    "specific": 1,
                    "firmware_version": "3.2",
                    "keep_awake": false,
                    "db_link": "https://devices.zwave-js.io/?jumpTo=0x010f:0x0602:0x1003:3.2",
                    "zwave_plus_version": 1,
                    "zwave_plus_node_type": 0,
                    "zwave_plus_role_type": 5,
                    "neighbours": [
                      1
                    ],
                    "groups": {
                      "1": {
                        "label": "Lifeline",
                        "index": 1,
                        "multichannel": true,
                        "max_associations": 1
                      },
                      "2": {
                        "label": "On/Off (Button)",
                        "index": 2,
                        "multichannel": true,
                        "max_associations": 10
                      },
                      "3": {
                        "label": "On/Off (Power)",
                        "index": 3,
                        "multichannel": true,
                        "max_associations": 10
                      }
                    },
                    "values": {
                      "4-37-0-currentValue": {
                        "id": "4-37-0-currentValue",
                        "id_on_network": "4-37-0-currentValue",
                        "value_id": "4-37-0-currentValue",
                        "data": true,
                        "data_items": null,
                        "label": "Current value",
                        "property_id": "currentValue",
                        "help": null,
                        "node_id": 4,
                        "parent_id": 4,
                        "type": "Bool",
                        "default": null,
                        "endpoint": 0,
                        "stateless": false,
                        "command_class": 37,
                        "command_class_name": "Binary Switch",
                        "units": null,
                        "min": null,
                        "max": null,
                        "is_read_only": true,
                        "is_write_only": false,
                        "last_update": "2021-04-05T19:38:07.587000"
                      }
                    }
                  }
                }

        :param kwargs: Extra arguments to be passed to :meth:`platypush.plugins.mqtt.MqttPlugin.publish``
            (default: query the default configured device).
        """
        if node_id or node_name:
            return self._get_node(node_id=node_id, node_name=node_name, use_cache=False, **kwargs)

        nodes = {
            node['id']: self.node_to_dict(node)
            for node in self._api_request('getNodes', **kwargs)
        }

        self._nodes_cache['by_id'] = nodes
        self._nodes_cache['by_name'] = {
            node['name']: node
            for node in nodes.values()
        }

        return nodes

    @action
    def get_node_stats(self, **kwargs):
        """
        Get the statistics of a node on the network (not implemented by zwavejs2mqtt).
        """
        raise _NOT_IMPLEMENTED_ERR

    @action
    def set_node_name(self, new_name: str, node_id: Optional[int] = None, node_name: Optional[str] = None, **kwargs):
        """
        Rename a node on the network.

        :param new_name: New name for the node.
        :param node_id: Filter by node_id.
        :param node_name: Filter by current node name.
        :param kwargs: Extra arguments to be passed to :meth:`platypush.plugins.mqtt.MqttPlugin.publish``
            (default: query the default configured device).
        """
        if node_name:
            node_id = self._get_node(node_name=node_name, **kwargs).get('node_id')

        assert node_id, f'No such node: {node_id}'
        self._api_request('setNodeName', node_id, new_name, **kwargs)
        get_bus().post(ZwaveNodeRenamedEvent(node={
            **self._get_node(node_id=node_id),
            'name': new_name,
        }))

    @action
    def set_node_product_name(self, **kwargs):
        """
        Set the product name of a node (not implemented by zwavejs2mqtt).
        """
        raise _NOT_IMPLEMENTED_ERR

    @action
    def set_node_manufacturer_name(self, **kwargs):
        """
        Set the manufacturer name of a node (not implemented by zwavejs2mqtt).
        """
        raise _NOT_IMPLEMENTED_ERR

    @action
    def set_node_location(self, location: str, node_id: Optional[int] = None, node_name: Optional[str] = None,
                          **kwargs):
        """
        Set the location of a node.

        :param location: Node location.
        :param node_id: Filter by node_id.
        :param node_name: Filter by current node name.
        :param kwargs: Extra arguments to be passed to :meth:`platypush.plugins.mqtt.MqttPlugin.publish``
            (default: query the default configured device).
        """
        if node_name:
            node_id = self._get_node(node_name=node_name, **kwargs).get('node_id')

        assert node_id, f'No such node: {node_id}'
        self._api_request('setNodeLocation', node_id, location, **kwargs)
        get_bus().post(ZwaveNodeEvent(node={
            **self._get_node(node_id=node_id),
            'location': location,
        }))

    @action
    def cancel_command(self, **kwargs):
        """
        Cancel the current running command (not implemented by zwavejs2mqtt).
        """
        raise _NOT_IMPLEMENTED_ERR

    @action
    def kill_command(self, **kwargs):
        """
        Immediately terminate any running command on the controller and release the lock (not implemented by
        zwavejs2mqtt).
        """
        raise _NOT_IMPLEMENTED_ERR

    @action
    def set_controller_name(self, **kwargs):
        """
        Set the name of the controller on the network (not implemented: use
        :meth:`platypush.plugin.zwave.mqtt.ZwaveMqttPlugin.set_node_name` instead).
        """
        raise _NOT_IMPLEMENTED_ERR

    @action
    def get_capabilities(self, **kwargs) -> List[str]:
        """
        Get the capabilities of the controller (not implemented: use
        :meth:`platypush.plugin.zwave.mqtt.ZwaveMqttPlugin.get_nodes` instead).
        """
        raise _NOT_IMPLEMENTED_ERR

    @action
    def receive_configuration(self, **kwargs):
        """
        Receive the configuration from the primary controller on the network. Requires a primary controller active
        (not implemented by zwavejs2mqtt).
        """
        raise _NOT_IMPLEMENTED_ERR

    @action
    def transfer_primary_role(self, **kwargs):
        """
        Add a new controller to the network and make it the primary.
        The existing primary will become a secondary controller (not implemented by zwavejs2mqtt).
        """
        raise _NOT_IMPLEMENTED_ERR

    @action
    def heal(self, timeout: Optional[int] = 60, **kwargs):
        """
        Heal network by requesting nodes rediscover their neighbours.

        :param timeout: Duration of the healing process in seconds (default: 60). Set to zero or null for no timeout.
        :param kwargs: Extra arguments to be passed to :meth:`platypush.plugins.mqtt.MqttPlugin.publish``
            (default: query the default configured device).
        """
        self._api_request('beginHealingNetwork', **kwargs)
        if timeout:
            Timer(timeout, lambda: self._api_request('stopHealingNetwork', **kwargs)).start()

    @action
    def switch_all(self, **kwargs):
        """
        Switch all the connected devices on/off (not implemented).
        """
        raise _NOT_IMPLEMENTED_ERR

    @action
    def test(self, count: int = 1, **kwargs):
        """
        Send a number of test messages to every node and record results (not implemented by zwavejs2mqtt).
        """
        raise _NOT_IMPLEMENTED_ERR

    @action
    def get_value(self, value_id: Optional[int] = None, id_on_network: Optional[str] = None,
                  value_label: Optional[str] = None, node_id: Optional[int] = None, node_name: Optional[str] = None,
                  **kwargs) -> Dict[str, Any]:
        """
        Get a value on the network.

        :param value_id: Select by value_id.
        :param id_on_network: Select value by id_on_network.
        :param value_label: Select value by [node_id/node_name, value_label]
        :param node_id: Select value by [node_id/node_name, value_label]
        :param node_name: Select value by [node_id/node_name, value_label]
        :param kwargs: Extra arguments to be passed to :meth:`platypush.plugins.mqtt.MqttPlugin.publish``
            (default: query the default configured device).
        """
        return self._get_value(value_id=value_id, value_label=value_label, id_on_network=id_on_network, node_id=node_id,
                               node_name=node_name, use_cache=False, **kwargs)

    @action
    def set_value(self, data, value_id: Optional[int] = None, id_on_network: Optional[str] = None,
                  value_label: Optional[str] = None, node_id: Optional[int] = None, node_name: Optional[str] = None,
                  **kwargs):
        """
        Set a value.

        :param data: Data to set for the value.
        :param value_id: Select value by value_id.
        :param id_on_network: Select value by id_on_network.
        :param value_label: Select value by [node_id/node_name, value_label]
        :param node_id: Select value by [node_id/node_name, value_label]
        :param node_name: Select value by [node_id/node_name, value_label]
        :param kwargs: Extra arguments to be passed to :meth:`platypush.plugins.mqtt.MqttPlugin.publish``
            (default: query the default configured device).
        """
        value = self._get_value(value_id=value_id, value_label=value_label, id_on_network=id_on_network,
                                node_id=node_id, node_name=node_name, **kwargs)

        self._api_request('writeValue', {
            'nodeId': value['node_id'],
            'commandClass': value['command_class'],
            'endpoint': value.get('endpoint', 0),
            'property': value['property_id'],
            **({'propertyKey': value['property_key']} if 'property_key' in value else {}),
        }, data, **kwargs)

    @action
    def set_value_label(self, **kwargs):
        """
        Change the label/name of a value (not implemented by zwavejs2mqtt).
        """
        raise _NOT_IMPLEMENTED_ERR

    @action
    def node_add_value(self, **kwargs):
        """
        Add a value to a node (not implemented by zwavejs2mqtt).
        """
        raise _NOT_IMPLEMENTED_ERR

    @action
    def node_remove_value(self, **kwargs):
        """
        Remove a value from a node (not implemented by zwavejs2mqtt).
        """
        raise _NOT_IMPLEMENTED_ERR

    @action
    def node_heal(self, node_id: Optional[int] = None, node_name: Optional[str] = None, **kwargs):
        """
        Heal network node by requesting the node to rediscover their neighbours.

        :param node_id: Select node by node_id.
        :param node_name: Select node by label.
        :param kwargs: Extra arguments to be passed to :meth:`platypush.plugins.mqtt.MqttPlugin.publish``
            (default: query the default configured device).
        """
        if node_name:
            node_id = self._get_node(node_name=node_name, **kwargs).get('node_id')
        self._api_request('healNode', node_id, **kwargs)

    @action
    def node_update_neighbours(self, node_id: Optional[int] = None, node_name: Optional[str] = None, **kwargs):
        """
        Ask a node to update its neighbours table
        (same as :meth:`platypush.plugins.zwave.mqtt.ZwaveMqttPlugin.node_heal`).

        :param node_id: Select node by node_id.
        :param node_name: Select node by label.
        :param kwargs: Extra arguments to be passed to :meth:`platypush.plugins.mqtt.MqttPlugin.publish``
            (default: query the default configured device).
        """
        self.node_heal(node_id=node_id, node_name=node_name, **kwargs)

    @action
    def node_network_update(self, node_id: Optional[int] = None, node_name: Optional[str] = None, **kwargs):
        """
        Update the controller with network information
        (same as :meth:`platypush.plugins.zwave.mqtt.ZwaveMqttPlugin.node_heal`).

        :param node_id: Select node by node_id.
        :param node_name: Select node by label.
        :param kwargs: Extra arguments to be passed to :meth:`platypush.plugins.mqtt.MqttPlugin.publish``
            (default: query the default configured device).
        """
        self.node_heal(node_id=node_id, node_name=node_name, **kwargs)

    @action
    def node_refresh_info(self, node_id: Optional[int] = None, node_name: Optional[str] = None, **kwargs):
        """
        Fetch up-to-date information about the node.

        :param node_id: Select node by node_id.
        :param node_name: Select node by label.
        :param kwargs: Extra arguments to be passed to :meth:`platypush.plugins.mqtt.MqttPlugin.publish``
            (default: query the default configured device).
        """
        if node_name:
            node_id = self._get_node(node_name=node_name, **kwargs).get('node_id')
        self._api_request('refreshInfo', node_id, **kwargs)

    @action
    def get_dimmers(self, node_id: Optional[int] = None, node_name: Optional[str] = None, **kwargs) -> Dict[str, Any]:
        """
        Get the dimmers on the network or associated to a node.

        :param node_id: Select node by node_id.
        :param node_name: Select node by label.
        :param kwargs: Extra arguments to be passed to :meth:`platypush.plugins.mqtt.MqttPlugin.publish``
            (default: query the default configured device).
        """
        return self._filter_values(['switch_multilevel', 'switch_toggle_multilevel'], node_id=node_id,
                                   node_name=node_name, **kwargs)

    @action
    def get_node_config(self, node_id: Optional[int] = None, node_name: Optional[str] = None,
                        **kwargs) -> Dict[str, Any]:
        """
        Get the configuration values of a node or of all the nodes on the network.

        :param node_id: Select node by node_id.
        :param node_name: Select node by label.
        :param kwargs: Extra arguments to be passed to :meth:`platypush.plugins.mqtt.MqttPlugin.publish``
            (default: query the default configured device).
        """
        return self._filter_values(['configuration', 'ip_configuration', 'association_command_configuration',
                                    'sensor_configuration'], node_id=node_id, node_name=node_name, **kwargs)

    @action
    def get_battery_levels(self, node_id: Optional[int] = None, node_name: Optional[str] = None,
                           **kwargs) -> Dict[str, Any]:
        """
        Get the battery levels of a node or of all the nodes on the network.

        :param node_id: Select node by node_id.
        :param node_name: Select node by name.
        :param kwargs: Extra arguments to be passed to :meth:`platypush.plugins.mqtt.MqttPlugin.publish``
            (default: query the default configured device).
        """
        return self._filter_values(['battery'], node_id=node_id, node_name=node_name, **kwargs)

    @action
    def get_power_levels(self, node_id: Optional[int] = None, node_name: Optional[str] = None,
                         **kwargs) -> Dict[str, Any]:
        """
        Get the power levels of this node.

        :param node_id: Select node by node_id.
        :param node_name: Select node by name.
        :param kwargs: Extra arguments to be passed to :meth:`platypush.plugins.mqtt.MqttPlugin.publish``
            (default: query the default configured device).
        """
        return self._filter_values(['powerlevel'], node_id=node_id, node_name=node_name, **kwargs)

    @action
    def get_bulbs(self, node_id: Optional[int] = None, node_name: Optional[str] = None, **kwargs) -> Dict[str, Any]:
        """
        Get the bulbs/LEDs on the network or associated to a node.

        :param node_id: Select node by node_id.
        :param node_name: Select node by name.
        :param kwargs: Extra arguments to be passed to :meth:`platypush.plugins.mqtt.MqttPlugin.publish``
            (default: query the default configured device).
        """
        return self._filter_values(['color'], filter_callback=lambda value: not value['is_read_only'], node_id=node_id,
                                   node_name=node_name, **kwargs)

    @action
    def get_switches(self, node_id: Optional[int] = None, node_name: Optional[str] = None, **kwargs) -> Dict[str, Any]:
        """
        Get the switches on the network or associated to a node.

        :param node_id: Select node by node_id.
        :param node_name: Select node by name.
        :param kwargs: Extra arguments to be passed to :meth:`platypush.plugins.mqtt.MqttPlugin.publish``
            (default: query the default configured device).
        """
        return self._filter_values(['switch_binary', 'switch_toggle_binary'],
                                   filter_callback=lambda value: not value['is_read_only'],
                                   node_id=node_id, node_name=node_name, **kwargs)

    @action
    def get_sensors(self, node_id: Optional[int] = None, node_name: Optional[str] = None, **kwargs) -> Dict[str, Any]:
        """
        Get the sensors on the network or associated to a node.

        :param node_id: Select node by node_id.
        :param node_name: Select node by name.
        :param kwargs: Extra arguments to be passed to :meth:`platypush.plugins.mqtt.MqttPlugin.publish``
            (default: query the default configured device).
        """
        return self._filter_values(['sensor_binary', 'sensor_multilevel', 'sensor_alarm', 'meter'],
                                   filter_callback=lambda value: not value['is_write_only'],
                                   node_id=node_id, node_name=node_name, **kwargs)

    @action
    def get_doorlocks(self, node_id: Optional[int] = None, node_name: Optional[str] = None, **kwargs) -> Dict[str, Any]:
        """
        Get the doorlocks on the network or associated to a node.

        :param node_id: Select node by node_id.
        :param node_name: Select node by name.
        :param kwargs: Extra arguments to be passed to :meth:`platypush.plugins.mqtt.MqttPlugin.publish``
            (default: query the default configured device).
        """
        return self._filter_values(['door_lock'], node_id=node_id, node_name=node_name, **kwargs)

    @action
    def get_locks(self, node_id: Optional[int] = None, node_name: Optional[str] = None, **kwargs) -> Dict[str, Any]:
        """
        Get the locks on the network or associated to a node.

        :param node_id: Select node by node_id.
        :param node_name: Select node by name.
        :param kwargs: Extra arguments to be passed to :meth:`platypush.plugins.mqtt.MqttPlugin.publish``
            (default: query the default configured device).
        """
        return self._filter_values(['lock'], node_id=node_id, node_name=node_name, **kwargs)

    @action
    def get_usercodes(self, node_id: Optional[int] = None, node_name: Optional[str] = None, **kwargs) -> Dict[str, Any]:
        """
        Get the usercodes on the network or associated to a node.

        :param node_id: Select node by node_id.
        :param node_name: Select node by name.
        :param kwargs: Extra arguments to be passed to :meth:`platypush.plugins.mqtt.MqttPlugin.publish``
            (default: query the default configured device).
        """
        return self._filter_values(['user_code'], node_id=node_id, node_name=node_name, **kwargs)

    @action
    def get_thermostats(self, node_id: Optional[int] = None, node_name: Optional[str] = None, **kwargs) \
            -> Dict[str, Any]:
        """
        Get the thermostats on the network or associated to a node.

        :param node_id: Select node by node_id.
        :param node_name: Select node by name.
        :param kwargs: Extra arguments to be passed to :meth:`platypush.plugins.mqtt.MqttPlugin.publish``
            (default: query the default configured device).
        """
        return self._filter_values(['thermostat_heating', 'thermostat_mode', 'thermostat_operating_state',
                                    'thermostat_setpoint', 'thermostat_fan_mode', 'thermostat_fan_state',
                                    'thermostat_setback'], node_id=node_id, node_name=node_name, **kwargs)

    @action
    def get_protections(self, node_id: Optional[int] = None, node_name: Optional[str] = None, **kwargs) \
            -> Dict[str, Any]:
        """
        Get the protection-compatible devices on the network or associated to a node.

        :param node_id: Select node by node_id.
        :param node_name: Select node by name.
        :param kwargs: Extra arguments to be passed to :meth:`platypush.plugins.mqtt.MqttPlugin.publish``
            (default: query the default configured device).
        """
        return self._filter_values(['protection'], node_id=node_id, node_name=node_name, **kwargs)

    @action
    def get_groups(self, **kwargs) -> Dict[str, dict]:
        """
        Get the groups on the network.

        :param kwargs: Extra arguments to be passed to :meth:`platypush.plugins.mqtt.MqttPlugin.publish``
            (default: query the default configured device).

        :return: A list of the available groups. Example:

          .. code-block:: json

            {
              "2-2": {
                "label": "Motion",
                "multichannel": true,
                "max_associations": 10,
                "group_id": "2-2",
                "node_id": 2,
                "index": 2,
                "associations": [
                  3,
                  4
                ]
              },
              "5-3": {
                "label": "Tamper",
                "multichannel": true,
                "max_associations": 10,
                "group_id": "5-3",
                "node_id": 5,
                "index": 3,
                "associations": [
                  2,
                  3,
                  4
                ]
              }
            ]

        """
        # noinspection PyUnresolvedReferences
        nodes = self.get_nodes(**kwargs).output
        self._groups_cache = {
            group['group_id']: {
                **group,
                'associations': [
                    assoc['nodeId'] for assoc in self._api_request('getAssociations', node_id, group['index'])
                ],
            }
            for node_id, node in nodes.items()
            for group in node.get('groups', {}).values()
        }

        return self._groups_cache

    @action
    def get_scenes(self, **kwargs) -> Dict[int, Dict[str, Any]]:
        """
        Get the scenes configured on the network.

        :param kwargs: Extra arguments to be passed to :meth:`platypush.plugins.mqtt.MqttPlugin.publish``
            (default: query the default configured device).
        :return: dict with the following format:

          .. code-block:: json

            {
              1: {
                "scene_id": 1,
                "label": "Scene 1",
                "values": {
                  "4-37-0-targetValue": {
                     "node_id": 4,
                     "value_id": "4-37-0-targetValue"
                     "data": true
                  },
                  "3-112-0-Temperature": {
                    "node_id": 3,
                    "value_id": "3-112-0-Temperature",
                    "data": true
                  }
                }
              }
            }

        """
        return {
            scene.get('sceneid'): self.scene_to_dict(scene)
            for scene in self._api_request('_getScenes')
        }

    @action
    def create_scene(self, label: str, **kwargs):
        """
        Create a new scene.

        :param label: Scene label.
        :param kwargs: Extra arguments to be passed to :meth:`platypush.plugins.mqtt.MqttPlugin.publish``
            (default: query the default configured device).
        """
        self._api_request('_createScene', label)

    @action
    def remove_scene(self, scene_id: Optional[int] = None, scene_label: Optional[str] = None, **kwargs):
        """
        Remove a scene.

        :param scene_id: Select by scene_id.
        :param scene_label: Select by scene label.
        :param kwargs: Extra arguments to be passed to :meth:`platypush.plugins.mqtt.MqttPlugin.publish``
            (default: query the default configured device).
        """
        scene = self._get_scene(scene_id=scene_id, scene_label=scene_label, **kwargs)
        self._api_request('_removeScene', scene['scene_id'])

    @action
    def activate_scene(self, scene_id: Optional[int] = None, scene_label: Optional[str] = None, **kwargs):
        """
        Activate a scene.

        :param scene_id: Select by scene_id.
        :param scene_label: Select by scene label.
        :param kwargs: Extra arguments to be passed to :meth:`platypush.plugins.mqtt.MqttPlugin.publish``
            (default: query the default configured device).
        """
        scene = self._get_scene(scene_id=scene_id, scene_label=scene_label, **kwargs)
        self._api_request('_activateScene', scene['scene_id'])

    @action
    def set_scene_label(self, new_label: str, scene_id: Optional[int] = None, scene_label: Optional[str] = None,
                        **kwargs):
        """
        Rename a scene/set the scene label.

        :param new_label: New label.
        :param scene_id: Select by scene_id.
        :param scene_label: Select by current scene label.
        :param kwargs: Extra arguments to be passed to :meth:`platypush.plugins.mqtt.MqttPlugin.publish``
            (default: query the default configured device).
        """
        raise _NOT_IMPLEMENTED_ERR

    @action
    def scene_add_value(self, data: Optional[Any] = None, value_id: Optional[int] = None,
                        id_on_network: Optional[str] = None, value_label: Optional[str] = None,
                        scene_id: Optional[int] = None, scene_label: Optional[str] = None,
                        node_id: Optional[int] = None, node_name: Optional[str] = None, **kwargs):
        """
        Add a value to a scene.

        :param data: Data to set for the value (default: current value data).
        :param value_id: Select value by value_id.
        :param id_on_network: Select value by id_on_network.
        :param value_label: Select value by [node_id/node_name, value_label]
        :param node_id: Select value by [node_id/node_name, value_label]
        :param node_name: Select value by [node_id/node_name, value_label]
        :param scene_id: Select scene by scene_id.
        :param scene_label: Select scene by scene label.
        :param kwargs: Extra arguments to be passed to :meth:`platypush.plugins.mqtt.MqttPlugin.publish``
            (default: query the default configured device).
        """
        value = self._get_value(value_id=value_id, value_label=value_label, id_on_network=id_on_network,
                                node_id=node_id, node_name=node_name, **kwargs)
        scene = self._get_scene(scene_id=scene_id, scene_label=scene_label, **kwargs)

        self._api_request('_addSceneValue', scene['scene_id'], {
            'nodeId': value['node_id'],
            'commandClass': value['command_class'],
            'property': value['property_id'],
            'endpoint': value['endpoint'],
        }, data, kwargs.get('timeout', self.timeout))

    @action
    def scene_remove_value(self, value_id: Optional[int] = None, id_on_network: Optional[str] = None,
                           value_label: Optional[str] = None, scene_id: Optional[int] = None,
                           scene_label: Optional[str] = None, node_id: Optional[int] = None,
                           node_name: Optional[str] = None, **kwargs):
        """
        Remove a value from a scene.

        :param value_id: Select value by value_id.
        :param id_on_network: Select value by id_on_network.
        :param value_label: Select value by [node_id/node_name, value_label]
        :param node_id: Select value by [node_id/node_name, value_label]
        :param node_name: Select value by [node_id/node_name, value_label]
        :param scene_id: Select scene by scene_id.
        :param scene_label: Select scene by scene label.
        :param kwargs: Extra arguments to be passed to :meth:`platypush.plugins.mqtt.MqttPlugin.publish``
            (default: query the default configured device).
        """
        value = self._get_value(value_id=value_id, value_label=value_label, id_on_network=id_on_network,
                                node_id=node_id, node_name=node_name, **kwargs)
        scene = self._get_scene(scene_id=scene_id, scene_label=scene_label, **kwargs)
        self._api_request('_removeSceneValue', scene['scene_id'], value['value_id'])

    @action
    def get_scene_values(self, scene_id: Optional[int] = None, scene_label: Optional[str] = None, **kwargs) -> dict:
        """
        Get the values associated to a scene.

        :param scene_id: Select by scene_id.
        :param scene_label: Select by scene label.
        :param kwargs: Extra arguments to be passed to :meth:`platypush.plugins.mqtt.MqttPlugin.publish``
            (default: query the default configured device).
        """
        scene = self._get_scene(scene_id=scene_id, scene_label=scene_label, **kwargs)
        return scene.get('values', {})

    @action
    def create_button(self, button_id: Union[int, str], node_id: Optional[int] = None, node_name: Optional[str] = None,
                      **kwargs):
        """
        Create a handheld button on a device. Only intended for bridge firmware controllers
        (not implemented by zwavejs2mqtt).

        :param button_id: The ID of the button.
        :param node_id: Filter by node_id.
        :param node_name: Filter by current node name.
        :param kwargs: Extra arguments to be passed to :meth:`platypush.plugins.mqtt.MqttPlugin.publish``
            (default: query the default configured device).
        """
        raise _NOT_IMPLEMENTED_ERR

    @action
    def delete_button(self, button_id: Union[int, str], node_id: Optional[int] = None, node_name: Optional[str] = None,
                      **kwargs):
        """
        Delete a button association from a device. Only intended for bridge firmware controllers.
        (not implemented by zwavejs2mqtt).

        :param button_id: The ID of the button.
        :param node_id: Filter by node_id.
        :param node_name: Filter by current node name.
        :param kwargs: Extra arguments to be passed to :meth:`platypush.plugins.mqtt.MqttPlugin.publish``
            (default: query the default configured device).
        """
        raise _NOT_IMPLEMENTED_ERR

    @action
    def add_node_to_group(self, group_id: Optional[str] = None, node_id: Optional[int] = None,
                          endpoint: Optional[int] = None, **kwargs):
        """
        Add a node to a group.

        :param group_id: Group ID.
        :param node_id: Node ID to be added.
        :param endpoint: Add a specific endpoint of the node to the group (default: add a node association).
        :param kwargs: Extra arguments to be passed to :meth:`platypush.plugins.mqtt.MqttPlugin.publish``
            (default: query the default configured device).
        """
        group = self._get_group(group_id, **kwargs)
        assoc = {'nodeId': node_id}
        if endpoint is not None:
            assoc['endpoint'] = endpoint

        self._api_request('addAssociations', group['node_id'], group['index'], [assoc])

    @action
    def remove_node_from_group(self, group_id: Optional[str] = None, node_id: Optional[int] = None,
                               endpoint: Optional[int] = None, **kwargs):
        """
        Remove a node from a group.

        :param group_id: Group ID.
        :param node_id: Node ID to be added.
        :param endpoint: Node endpoint to remove (default: remove node association).
        :param kwargs: Extra arguments to be passed to :meth:`platypush.plugins.mqtt.MqttPlugin.publish``
            (default: query the default configured device).
        """
        group = self._get_group(group_id, **kwargs)
        assoc = {'nodeId': node_id}
        if endpoint is not None:
            assoc['endpoint'] = endpoint

        self._api_request('removeAssociations', group['node_id'], group['index'], [assoc])

    @action
    def create_new_primary(self, **kwargs):
        """
        Create a new primary controller on the network when the previous primary fails
        (not implemented by zwavejs2mqtt).

        :param kwargs: Extra arguments to be passed to :meth:`platypush.plugins.mqtt.MqttPlugin.publish``
            (default: query the default configured device).
        """
        raise _NOT_IMPLEMENTED_ERR

    @action
    def hard_reset(self, **kwargs):
        """
        Perform a hard reset of the controller. It erases its network configuration settings.
        The controller becomes a primary controller ready to add devices to a new network.

        :param kwargs: Extra arguments to be passed to :meth:`platypush.plugins.mqtt.MqttPlugin.publish``
            (default: query the default configured device).
        """
        self._api_request('hardReset')

    @action
    def soft_reset(self, **kwargs):
        """
        Perform a soft reset of the controller.
        Resets a controller without erasing its network configuration settings (not implemented by zwavejs2).

        :param kwargs: Extra arguments to be passed to :meth:`platypush.plugins.mqtt.MqttPlugin.publish``
            (default: query the default configured device).
        """
        raise _NOT_IMPLEMENTED_ERR

    @action
    def write_config(self, **kwargs):
        """
        Store the current configuration of the network to the user directory (not implemented by zwavejs2mqtt).

        :param kwargs: Extra arguments to be passed to :meth:`platypush.plugins.mqtt.MqttPlugin.publish``
            (default: query the default configured device).
        """
        raise _NOT_IMPLEMENTED_ERR

    @action
    def on(self, device: str, *args, **kwargs):
        """
        Turn on a switch on a device.

        :param device: ``id_on_network`` of the value to be switched on.
        :param kwargs: Extra arguments to be passed to :meth:`platypush.plugins.mqtt.MqttPlugin.publish``
            (default: query the default configured device).
        """
        self.set_value(data=True, id_on_network=device, **kwargs)

    @action
    def off(self, device: str, *args, **kwargs):
        """
        Turn off a switch on a device.

        :param device: ``id_on_network`` of the value to be switched off.
        :param kwargs: Extra arguments to be passed to :meth:`platypush.plugins.mqtt.MqttPlugin.publish``
            (default: query the default configured device).
        """
        self.set_value(data=False, id_on_network=device, **kwargs)

    @action
    def toggle(self, device: str, *args, **kwargs) -> dict:
        """
        Toggle a switch on a device.

        :param device: ``id_on_network`` of the value to be toggled.
        :param kwargs: Extra arguments to be passed to :meth:`platypush.plugins.mqtt.MqttPlugin.publish``
            (default: query the default configured device).
        """
        value = self._get_value(id_on_network=device, use_cache=False, **kwargs)
        value['data'] = not value['data']
        self.set_value(data=value['data'], id_on_network=device, **kwargs)

        return {
            'name': '{} - {}'.format(self._nodes_cache['by_id'][value['node_id']]['name'],
                                     value.get('label', '[No Label]')),
            'on': value['data'],
            'id': value['value_id'],
        }

    @property
    def switches(self) -> List[dict]:
        # Repopulate the nodes cache
        self.get_nodes()
        # noinspection PyUnresolvedReferences
        devices = self.get_switches().output.values()
        return [
            {
                'name': '{} - {}'.format(self._nodes_cache['by_id'][dev['node_id']]['name'],
                                         dev.get('label', '[No Label]')),
                'on': dev['data'],
                'id': dev['value_id'],
            } for dev in devices
        ]


# vim:sw=4:ts=4:et:



from platypush.plugins import Plugin, action


class MusicPlugin(Plugin):
    def __init__(self, *args, **kwargs):
        super().__init__(**kwargs)

    @action
    def play(self, **kwargs):
        raise NotImplementedError()

    @action
    def pause(self, **kwargs):
        raise NotImplementedError()

    @action
    def stop(self, **kwargs):
        raise NotImplementedError()

    @action
    def next(self, **kwargs):
        raise NotImplementedError()

    @action
    def previous(self, **kwargs):
        raise NotImplementedError()

    @action
    def set_volume(self, volume, **kwargs):
        raise NotImplementedError()

    @action
    def volup(self, delta, **kwargs):
        raise NotImplementedError()

    @action
    def voldown(self, delta, **kwargs):
        raise NotImplementedError()

    @action
    def seek(self, position, **kwargs):
        raise NotImplementedError()

    @action
    def add(self, resource, **kwargs):
        raise NotImplementedError()

    @action
    def clear(self, **kwargs):
        raise NotImplementedError()

    @action
    def status(self, **kwargs):
        raise NotImplementedError()

    @action
    def current_track(self, **kwargs):
        raise NotImplementedError()

    @action
    def get_playlists(self, **kwargs):
        raise NotImplementedError()

    @action
    def get_playlist(self, playlist, **kwargs):
        raise NotImplementedError()

    @action
    def add_to_playlist(self, playlist, resources, **kwargs):
        raise NotImplementedError()

    @action
    def remove_from_playlist(self, playlist, resources, **kwargs):
        raise NotImplementedError()

    @action
    def playlist_move(self, playlist, from_pos: int, to_pos: int, **kwargs):
        raise NotImplementedError()

    @action
    def search(self, query, *args, **kwargs):
        raise NotImplementedError()


# vim:sw=4:ts=4:et:



import os
from typing import Sequence, Dict, Tuple, Union, Optional

from platypush.plugins import RunnablePlugin, action
from platypush.schemas.irc import IRCServerSchema, IRCServerStatusSchema, IRCChannelSchema

from ._bot import IRCBot
from .. import ChatPlugin


class ChatIrcPlugin(RunnablePlugin, ChatPlugin):
    """
    IRC integration.

    This plugin allows you to easily create IRC bots with custom logic that reacts to IRC events
    and interact with IRC sessions.

    Triggers:

        * :class:`platypush.message.event.irc.IRCChannelJoinEvent` when a user joins a channel.
        * :class:`platypush.message.event.irc.IRCChannelKickEvent` when a user is kicked from a channel.
        * :class:`platypush.message.event.irc.IRCModeEvent` when a user/channel mode change event occurs.
        * :class:`platypush.message.event.irc.IRCPartEvent` when a user parts a channel.
        * :class:`platypush.message.event.irc.IRCQuitEvent` when a user quits.
        * :class:`platypush.message.event.irc.IRCNickChangeEvent` when a user nick changes.
        * :class:`platypush.message.event.irc.IRCConnectEvent` when the bot connects to a server.
        * :class:`platypush.message.event.irc.IRCDisconnectEvent` when the bot disconnects from a server.
        * :class:`platypush.message.event.irc.IRCPrivateMessageEvent` when a private message is received.
        * :class:`platypush.message.event.irc.IRCPublicMessageEvent` when a public message is received.
        * :class:`platypush.message.event.irc.IRCDCCRequestEvent` when a DCC connection request is received.
        * :class:`platypush.message.event.irc.IRCDCCMessageEvent` when a DCC message is received.
        * :class:`platypush.message.event.irc.IRCCTCPMessageEvent` when a CTCP message is received.
        * :class:`platypush.message.event.irc.IRCDCCFileRequestEvent` when a DCC file request is received.
        * :class:`platypush.message.event.irc.IRCDCCFileRecvCompletedEvent` when a DCC file download is completed.
        * :class:`platypush.message.event.irc.IRCDCCFileRecvCancelledEvent` when a DCC file download is cancelled.
        * :class:`platypush.message.event.irc.IRCDCCFileSendCompletedEvent` when a DCC file upload is completed.
        * :class:`platypush.message.event.irc.IRCDCCFileSendCancelledEvent` when a DCC file upload is cancelled.

    Requires:

        * **irc** (``pip install irc``)

    """

    def __init__(self, servers: Sequence[dict], **kwargs):
        """
        :param servers: List of servers/channels that the bot will automatically connect/join.
        """
        super().__init__(**kwargs)
        try:
            self._bots: Dict[Tuple[str, int], IRCBot] = {
                (server_conf['server'], server_conf['port']): IRCBot(**server_conf)
                for server_conf in IRCServerSchema().load(servers, many=True)
            }
        except Exception as e:
            self.logger.warning(f'Could not load IRC server configuration: {e}')
            self.logger.exception(e)
            raise e

    @property
    def _bots_by_server(self) -> Dict[str, IRCBot]:
        return {
            bot.server: bot
            for srv, bot in self._bots.items()
        }

    @property
    def _bots_by_server_and_port(self) -> Dict[Tuple[str, int], IRCBot]:
        return {
            (bot.server, bot.port): bot
            for srv, bot in self._bots.items()
        }

    @property
    def _bots_by_alias(self) -> Dict[str, IRCBot]:
        return {
            bot.alias: bot
            for srv, bot in self._bots.items()
            if bot.alias
        }

    def main(self):
        self._connect()
        self._should_stop.wait()

    def _connect(self):
        for srv, bot in self._bots.items():
            self.logger.info(f'Connecting to IRC server {srv}')
            bot.start()

    def stop(self):
        for srv, bot in self._bots.items():
            self.logger.info(f'Disconnecting from IRC server {srv}')
            try:
                bot.stop(bot.stop_message or 'Application stopped')
            except Exception as e:
                self.logger.warning(f'Error while stopping connection to {srv}: {e}')

        super().stop()

    def _get_bot(self, server: Union[str, Tuple[str, int]]) -> IRCBot:
        if isinstance(server, (tuple, list, set)):
            bot = self._bots_by_server_and_port[tuple(server)]
        else:
            bot = self._bots_by_alias.get(server, self._bots_by_server.get(server))

        assert bot, f'Bot connection to {server} not found'
        return bot

    @action
    def send_file(
            self, file: str, server: Union[str, Tuple[str, int]], nick: str, bind_address: Optional[str] = None
    ):
        """
        Send a file to an IRC user over DCC connection.
        Note that passive connections are currently not supported.

        :param file: Path of the file that should be transferred.
        :param server: IRC server, identified either by ``alias`` or ``(server, port)`` tuple.
        :param nick: Target IRC nick.
        :param bind_address: DCC listen bind address (default: any).
        """
        file = os.path.expanduser(file)
        assert os.path.isfile(file), f'{file} is not a regular file'
        bot = self._get_bot(server)
        bot.dcc_file_transfer(file=file, nick=nick, bind_address=bind_address)

    @action
    def send_message(
            self, text: str, server: Union[str, Tuple[str, int]], target: Union[str, Sequence[str]]
    ):
        """
        Send a message to a channel or a nick.

        :param text: Message content.
        :param server: IRC server, identified either by ``alias`` or ``(server, port)`` tuple.
        :param target: Message target (nick or channel). If it's a list then the message will be sent
            to multiple targets.
        """
        bot = self._get_bot(server)
        method = (
            bot.connection.privmsg if isinstance(target, str)
            else bot.connection.privmsg_many
        )
        method(target, text)

    @action
    def send_notice(
            self, text: str, server: Union[str, Tuple[str, int]], target: str
    ):
        """
        Send a notice to a channel or a nick.

        :param text: Message content.
        :param server: IRC server, identified either by ``alias`` or ``(server, port)`` tuple.
        :param target: Message target (nick or channel).
        """
        bot = self._get_bot(server)
        bot.connection.notice(target, text)

    @action
    def servers(self) -> Sequence[dict]:
        """
        Get information about the connected servers.

        :return: .. schema:: irc.IRCServerStatusSchema(many=True)
        """
        bots = self._bots_by_server.values()
        return IRCServerStatusSchema().dump(
            {
                'server': bot.server,
                'port': bot.port,
                'alias': bot.alias,
                'real_name': bot.connection.get_server_name(),
                'nickname': bot.connection.get_nickname(),
                'is_connected': bot.connection.is_connected(),
                'connected_channels': bot.channels.keys(),
            }
            for bot in bots
        )

    @action
    def channel(self, server: Union[str, Tuple[str, int]], channel: str) -> dict:
        """
        Get information about a connected channel.

        :param server: IRC server, identified either by ``alias`` or ``(server, port)`` tuple.
        :param channel:
        :return: .. schema:: irc.IRCChannelSchema
        """
        bot = self._get_bot(server)
        channel_name = channel
        channel = bot.channels.get(channel)
        assert channel, f'Not connected to channel {channel}'
        return IRCChannelSchema().dump({
            'is_invite_only': channel.is_invite_only(),
            'is_moderated': channel.is_moderated(),
            'is_protected': channel.is_protected(),
            'is_secret': channel.is_secret(),
            'name': channel_name,
            'modes': channel.modes,
            'opers': list(channel.opers()),
            'owners': channel.owners(),
            'users': list(channel.users()),
            'voiced': list(channel.voiced()),
        })

    @action
    def send_ctcp_message(
            self, ctcp_type: str, body: str, server: Union[str, Tuple[str, int]], target: str
    ):
        """
        Send a CTCP message to a target.

        :param ctcp_type: CTCP message type.
        :param body: Message content.
        :param server: IRC server, identified either by ``alias`` or ``(server, port)`` tuple.
        :param target: Message target.
        """
        bot = self._get_bot(server)
        bot.connection.ctcp(ctcp_type, target, body)

    @action
    def send_ctcp_reply(
            self, body: str, server: Union[str, Tuple[str, int]], target: str
    ):
        """
        Send a CTCP REPLY command.

        :param body: Message content.
        :param server: IRC server, identified either by ``alias`` or ``(server, port)`` tuple.
        :param target: Message target.
        """
        bot = self._get_bot(server)
        bot.connection.ctcp_reply(target, body)

    @action
    def disconnect(self, server: Union[str, Tuple[str, int]], message: Optional[str] = None):
        """
        Disconnect from a server.

        :param server: IRC server, identified either by ``alias`` or ``(server, port)`` tuple.
        :param message: Disconnect message (default: configured ``stop_message``.
        """
        bot = self._get_bot(server)
        bot.connection.disconnect(message or bot.stop_message)

    @action
    def invite(
            self, nick: str, channel: str, server: Union[str, Tuple[str, int]]
    ):
        """
        Invite a nick to a channel.

        :param nick: Target IRC nick.
        :param channel: Target IRC channel.
        :param server: IRC server, identified either by ``alias`` or ``(server, port)`` tuple.
        """
        bot = self._get_bot(server)
        bot.connection.invite(nick, channel)

    @action
    def join(self, channel: str, server: Union[str, Tuple[str, int]]):
        """
        Join a channel.

        :param channel: Target IRC channel.
        :param server: IRC server, identified either by ``alias`` or ``(server, port)`` tuple.
        """
        bot = self._get_bot(server)
        bot.connection.join(channel)

    @action
    def kick(
            self, nick: str, channel: str, server: Union[str, Tuple[str, int]], reason: Optional[str] = None
    ):
        """
        Kick a nick from a channel.

        :param nick: Target IRC nick.
        :param channel: Target IRC channel.
        :param server: IRC server, identified either by ``alias`` or ``(server, port)`` tuple.
        :param reason: Kick reason.
        """
        bot = self._get_bot(server)
        bot.connection.kick(channel, nick, reason)

    @action
    def mode(
            self, target: str, command: str, server: Union[str, Tuple[str, int]]
    ):
        """
        Send a MODE command on the selected target.

        :param target: IRC target.
        :param command: Mode command.
        :param server: IRC server, identified either by ``alias`` or ``(server, port)`` tuple.
        """
        bot = self._get_bot(server)
        bot.connection.mode(target, command)

    @action
    def set_nick(self, nick: str, server: Union[str, Tuple[str, int]]):
        """
        Set the IRC nick.

        :param nick: New IRC nick.
        :param server: IRC server, identified either by ``alias`` or ``(server, port)`` tuple.
        """
        bot = self._get_bot(server)
        bot.connection.nick(nick)

    @action
    def oper(self, nick: str, password: str, server: Union[str, Tuple[str, int]]):
        """
        Send an OPER command.

        :param nick: IRC nick.
        :param password: Nick password.
        :param server: IRC server, identified either by ``alias`` or ``(server, port)`` tuple.
        """
        bot = self._get_bot(server)
        bot.connection.oper(nick, password)

    @action
    def part(
            self, channel: Union[str, Sequence[str]], server: Union[str, Tuple[str, int]],
            message: Optional[str] = None
    ):
        """
        Parts/exits a channel.

        :param channel: IRC channel (or list of channels).
        :param server: IRC server, identified either by ``alias`` or ``(server, port)`` tuple.
        :param message: Optional part message (default: same as the bot's ``stop_message``).
        """
        bot = self._get_bot(server)
        channels = [channel] if isinstance(channel, str) else channel
        bot.connection.part(channels=channels, message=message or bot.stop_message)

    @action
    def quit(
            self, server: Union[str, Tuple[str, int]], message: Optional[str] = None
    ):
        """
        Send a QUIT command.

        :param server: IRC server, identified either by ``alias`` or ``(server, port)`` tuple.
        :param message: Optional quit message (default: same as the bot's ``stop_message``).
        """
        bot = self._get_bot(server)
        bot.connection.quit(message=message or bot.stop_message)

    @action
    def send_raw(self, message: str, server: Union[str, Tuple[str, int]]):
        """
        Send a raw IRC message to a connected server.

        :param message: IRC message.
        :param server: IRC server, identified either by ``alias`` or ``(server, port)`` tuple.
        """
        bot = self._get_bot(server)
        bot.connection.send_raw(message)

    @action
    def topic(self, channel: str, server: Union[str, Tuple[str, int]], topic: Optional[str] = None) -> str:
        """
        Get/set the topic of an IRC channel.

        :param channel: IRC channel.
        :param server: IRC server, identified either by ``alias`` or ``(server, port)`` tuple.
        :param topic: If specified, then set the new topic as channel topic.
            Otherwise, just return the current channel topic.
        """
        bot = self._get_bot(server)
        with bot.event_queue('currenttopic') as evt_queue:
            bot.connection.topic(channel, topic)
            evt = evt_queue.get(block=True, timeout=bot.response_timeout)
            return evt.arguments[1]

    @action
    def whois(self, target: str, server: Union[str, Tuple[str, int]]):
        """
        Send a WHOIS command for a target.

        :param target: IRC target.
        :param server: IRC server, identified either by ``alias`` or ``(server, port)`` tuple.
        """
        bot = self._get_bot(server)
        with bot.event_queue('whoisuser') as evt_queue:
            bot.connection.whois([target])
            evt = evt_queue.get(block=True, timeout=bot.response_timeout)
            return evt.arguments


# vim:sw=4:ts=4:et:



import time

from platypush.backend import Backend
from platypush.context import get_plugin
from platypush.message.event.weather import NewWeatherConditionEvent


class WeatherBackend(Backend):
    """
    Abstract class for weather update backends.
    """

    def __init__(self, plugin_name: str, poll_seconds: int, **kwargs):
        """
        :param plugin_name: Name of the weather plugin to be used.
        :param poll_seconds: How often the backend should check for updates, in seconds.
        """
        super().__init__(**kwargs)
        self.plugin_name = plugin_name
        self.poll_seconds = poll_seconds
        self.latest_update = {}

    def run(self):
        super().run()
        weather = get_plugin(self.plugin_name)
        self.logger.info('Initialized {} backend'.format(self.__class__.__name__))

        while not self.should_stop():
            current_weather = weather.get_current_weather().output
            current_weather.pop('time', None)

            if current_weather != self.latest_update:
                self.bus.post(NewWeatherConditionEvent(plugin_name=self.plugin_name, **current_weather))

            self.latest_update = current_weather
            time.sleep(self.poll_seconds)


# vim:sw=4:ts=4:et:



from marshmallow import fields, INCLUDE
from marshmallow.schema import Schema


class MapsDistanceSchema(Schema):
    text = fields.String(required=True, metadata=dict(
        description='Distance expressed as readable text',
        example='6.5 km',
    ))

    value = fields.Number(required=True, metadata=dict(
        description='Distance expressed as a numeric value according to the selected units',
        example=6542,
    ))


class MapsDurationSchema(Schema):
    text = fields.String(required=True, metadata=dict(
        description='Duration expressed as readable text',
        example='13 mins',
    ))

    value = fields.Number(required=True, metadata=dict(
        description='Duration expressed in seconds',
        example=777,
    ))


class MapsTravelTimeSchema(Schema):
    class Meta:
        unknown = INCLUDE

    distance = fields.Nested(MapsDistanceSchema)
    duration = fields.Nested(MapsDurationSchema)



import asyncio
import json
import time

from typing import Optional, Collection

from websockets import connect as websocket_connect  # type: ignore
from websockets.exceptions import ConnectionClosed

from platypush.context import get_bus
from platypush.message.event.websocket import WebsocketMessageEvent
from platypush.plugins import AsyncRunnablePlugin, action
from platypush.utils import get_ssl_client_context


class WebsocketPlugin(AsyncRunnablePlugin):
    """
    Plugin to send and receive messages over websocket connections.

    Triggers:

        * :class:`platypush.message.event.websocket.WebsocketMessageEvent` when
            a message is received on a subscribed websocket.

    """

    def __init__(self, subscriptions: Optional[Collection[str]] = None, **kwargs):
        """
        :param subscriptions: List of websocket URLs that should be subscribed
            at startup, prefixed by ``ws://`` or ``wss://``.
        """
        super().__init__(**kwargs)
        self._subscriptions = subscriptions or []

    @property
    def loop(self):
        if not self._loop:
            self._loop = asyncio.new_event_loop()
            asyncio.set_event_loop(self._loop)

        return self._loop

    @action
    def send(
        self,
        url: str,
        msg,
        ssl_cert=None,
        ssl_key=None,
        ssl_cafile=None,
        ssl_capath=None,
        wait_response=False,
    ):
        """
        Sends a message to a websocket.

        :param url: Websocket URL, e.g. ws://localhost:8765 or wss://localhost:8765
        :param msg: Message to be sent. It can be a list, a dict, or a Message object
        :param ssl_cert: Path to the SSL certificate to be used, if the SSL
            connection requires client authentication as well (default: None)
        :param ssl_key: Path to the SSL key to be used, if the SSL connection
            requires client authentication as well (default: None)
        :param ssl_cafile: Path to the certificate authority file if required
            by the SSL configuration (default: None)
        :param ssl_capath: Path to the certificate authority directory if
            required by the SSL configuration (default: None)
        :param wait_response: Set to True if you expect a response to the
            delivered message.
        :return: The received response if ``wait_response`` is set to True,
            otherwise nothing.
        """

        msg = self._parse_msg(msg)

        async def send():
            websocket_args = {
                'ssl': self._get_ssl_context(
                    url,
                    ssl_cert=ssl_cert,
                    ssl_key=ssl_key,
                    ssl_cafile=ssl_cafile,
                    ssl_capath=ssl_capath,
                )
            }

            async with websocket_connect(url, **websocket_args) as ws:
                try:
                    await ws.send(str(msg))
                except ConnectionClosed as err:
                    self.logger.warning('Error on websocket %s: %s', url, err)

                if wait_response:
                    messages = await self._recv(ws, num_messages=1)
                    if messages:
                        return self._parse_msg(messages[0])

        return asyncio.run_coroutine_threadsafe(send(), self.loop).result()

    @action
    def recv(
        self,
        url: str,
        ssl_cert=None,
        ssl_key=None,
        ssl_cafile=None,
        ssl_capath=None,
        num_messages=0,
        timeout=0,
    ):
        """
        Receive one or more messages from a websocket.

        A :class:`platypush.message.event.websocket.WebsocketMessageEvent`
        event will be triggered whenever a new message is received.

        :param url: Websocket URL, e.g. ws://localhost:8765 or wss://localhost:8765
        :param ssl_cert: Path to the SSL certificate to be used, if the SSL
            connection requires client authentication as well (default: None)
        :param ssl_key: Path to the SSL key to be used, if the SSL connection
            requires client authentication as well (default: None)
        :param ssl_cafile: Path to the certificate authority file if required
            by the SSL configuration (default: None)
        :param ssl_capath: Path to the certificate authority directory if
            required by the SSL configuration (default: None)
        :param num_messages: Exit after receiving this number of messages.
            Default: 0, receive forever.
        :param timeout: Message receive timeout in seconds. Default: 0 - no timeout.
        :return: A list with the messages that have been received, unless
            ``num_messages`` is set to 0 or ``None``.
        """

        async def recv():
            websocket_args = {
                'ssl': self._get_ssl_context(
                    url,
                    ssl_cert=ssl_cert,
                    ssl_key=ssl_key,
                    ssl_cafile=ssl_cafile,
                    ssl_capath=ssl_capath,
                )
            }

            async with websocket_connect(url, **websocket_args) as ws:
                return await self._recv(ws, timeout=timeout, num_messages=num_messages)

        return self.loop.call_soon_threadsafe(recv)

    async def _recv(self, ws, timeout=0, num_messages=0):
        messages = []
        time_start = time.time()
        time_end = time_start + timeout if timeout else 0
        url = 'ws{secure}://{host}:{port}{path}'.format(
            secure='s' if ws._secure else '',
            host=ws.remote_address[0],
            port=ws.remote_address[1],
            path=ws.path,
        )

        while (not num_messages) or (len(messages) < num_messages):
            msg = None
            err = None
            remaining_timeout = time_end - time.time() if time_end else None

            try:
                msg = await asyncio.wait_for(ws.recv(), remaining_timeout)
            except (ConnectionClosed, asyncio.exceptions.TimeoutError) as e:
                err = e
                self.logger.warning('Error on websocket %s: %s', url, e)

            if isinstance(err, ConnectionClosed) or (
                time_end and time.time() > time_end
            ):
                break

            if msg is None:
                continue

            msg = self._parse_msg(msg)
            messages.append(msg)
            get_bus().post(WebsocketMessageEvent(url=url, message=msg))

        return messages

    @property
    def _should_start_runner(self):
        return bool(self._subscriptions)

    @staticmethod
    def _parse_msg(msg):
        try:
            msg = json.dumps(msg)
        except Exception:
            pass

        return msg

    async def listen(self):
        async def _recv(url):
            async with websocket_connect(url) as ws:
                return await self._recv(ws)

        await asyncio.wait([_recv(url) for url in set(self._subscriptions)])

    @staticmethod
    def _get_ssl_context(
        url: str, ssl_cert=None, ssl_key=None, ssl_cafile=None, ssl_capath=None
    ):
        if url.startswith('wss://') or url.startswith('https://'):
            return get_ssl_client_context(
                ssl_cert=ssl_cert,
                ssl_key=ssl_key,
                ssl_cafile=ssl_cafile,
                ssl_capath=ssl_capath,
            )

        return None


# vim:sw=4:ts=4:et:



from typing import Optional, Union

from platypush.exceptions import PlatypushException


class UserException(PlatypushException):
    """
    Base class for all user exceptions.
    """
    def __init__(self, *args, user: Optional[Union[str, int]] = None, **kwargs):
        super().__init__(*args, **kwargs)
        self.user = user


class AuthenticationException(UserException):
    """
    Authentication error exception.
    """
    def __init__(self, error='Unauthorized', *args, **kwargs):
        super().__init__(error, *args, **kwargs)


class InvalidTokenException(AuthenticationException):
    """
    Exception raised in case of wrong user token.
    """
    def __init__(self, error='Invalid user token', *args, **kwargs):
        super().__init__(error, *args, **kwargs)


class InvalidCredentialsException(AuthenticationException):
    """
    Exception raised in case of wrong user token.
    """
    def __init__(self, error='Invalid credentials', *args, **kwargs):
        super().__init__(error, *args, **kwargs)


class InvalidJWTTokenException(InvalidTokenException):
    """
    Exception raised in case of wrong/invalid JWT token.
    """
    def __init__(self, error='Invalid JWT token', *args, **kwargs):
        super().__init__(error, *args, **kwargs)



import json
from queue import Queue, Empty
from typing import Optional, Type

from platypush.backend.mqtt import MqttBackend
from platypush.context import get_plugin

from platypush.message.event.zwave import ZwaveEvent, ZwaveNodeAddedEvent, ZwaveValueChangedEvent, \
    ZwaveNodeRemovedEvent, ZwaveNodeRenamedEvent, ZwaveNodeReadyEvent, ZwaveNodeEvent, ZwaveNodeAsleepEvent, \
    ZwaveNodeAwakeEvent


class ZwaveMqttBackend(MqttBackend):
    """
    Listen for events on a `zwavejs2mqtt <https://github.com/zwave-js/zwavejs2mqtt>`_ service.

    Triggers:

        * :class:`platypush.message.event.zwave.ZwaveNodeEvent` when a node attribute changes.
        * :class:`platypush.message.event.zwave.ZwaveNodeAddedEvent` when a node is added to the network.
        * :class:`platypush.message.event.zwave.ZwaveNodeRemovedEvent` when a node is removed from the network.
        * :class:`platypush.message.event.zwave.ZwaveNodeRenamedEvent` when a node is renamed.
        * :class:`platypush.message.event.zwave.ZwaveNodeReadyEvent` when a node is ready.
        * :class:`platypush.message.event.zwave.ZwaveValueChangedEvent` when the value of a node on the network
          changes.
        * :class:`platypush.message.event.zwave.ZwaveNodeAsleepEvent` when a node goes into sleep mode.
        * :class:`platypush.message.event.zwave.ZwaveNodeAwakeEvent` when a node goes back into awake mode.

    Requires:

        * **paho-mqtt** (``pip install paho-mqtt``)
        * A `zwavejs2mqtt instance <https://github.com/zwave-js/zwavejs2mqtt>`_.
        * The :class:`platypush.plugins.zwave.mqtt.ZwaveMqttPlugin` plugin configured.

    """

    def __init__(self, client_id: Optional[str] = None, *args, **kwargs):
        """
        :param client_id: MQTT client ID (default: ``<device_id>-zwavejs-mqtt``, to prevent clashes with the
            :class:`platypush.backend.mqtt.MqttBackend` ``client_id``.
        """

        from platypush.plugins.zwave.mqtt import ZwaveMqttPlugin
        self.plugin: ZwaveMqttPlugin = get_plugin('zwave.mqtt')
        assert self.plugin, 'The zwave.mqtt plugin is not configured'

        self._nodes = {}
        self._groups = {}
        self._last_state = None
        self._events_queue = Queue()
        self.events_topic = self.plugin.events_topic
        self.server_info = {
            'host': self.plugin.host,
            'port': self.plugin.port or self._default_mqtt_port,
            'tls_cafile': self.plugin.tls_cafile,
            'tls_certfile': self.plugin.tls_certfile,
            'tls_ciphers': self.plugin.tls_ciphers,
            'tls_keyfile': self.plugin.tls_keyfile,
            'tls_version': self.plugin.tls_version,
            'username': self.plugin.username,
            'password': self.plugin.password,
        }

        listeners = [{
            **self.server_info,
            'topics': [
                self.plugin.events_topic + '/node/' + topic
                for topic in ['node_ready', 'node_sleep', 'node_value_updated', 'node_metadata_updated', 'node_wakeup']
            ],
        }]

        super().__init__(*args, subscribe_default_topic=False, listeners=listeners, client_id=client_id, **kwargs)
        if not client_id:
            self.client_id += '-zwavejs-mqtt'

    def _dispatch_event(self, event_type: Type[ZwaveEvent], node: Optional[dict] = None, value: Optional[dict] = None,
                        **kwargs):
        if value and 'id' not in value:
            value_id = f"{value['commandClass']}-{value.get('endpoint', 0)}-{value['property']}"
            if 'propertyKey' in value:
                value_id += '-' + str(value['propertyKey'])

            if value_id not in node.get('values', {}):
                self.logger.warning(f'value_id {value_id} not found on node {node["id"]}')
                return

            value = node['values'][value_id]

        if value:
            kwargs['value'] = self.plugin.value_to_dict(value)

        if node:
            kwargs['node'] = self.plugin.node_to_dict(node)
            node_id = kwargs['node']['node_id']

            if event_type == ZwaveNodeEvent:
                if node_id not in self._nodes:
                    event_type = ZwaveNodeAddedEvent
                elif kwargs['node']['name'] != self._nodes[node_id]['name']:
                    event_type = ZwaveNodeRenamedEvent

            if event_type == ZwaveNodeRemovedEvent:
                self._nodes.pop(node_id, None)
            else:
                self._nodes[node_id] = kwargs['node']

        evt = event_type(**kwargs)
        self._events_queue.put(evt)

        # zwavejs2mqtt currently treats some values (e.g. binary switches) in an inconsistent way,
        # using two values - a read-only value called currentValue that gets updated on the
        # node_value_updated topic, and a writable value called targetValue that doesn't get updated
        # (see https://github.com/zwave-js/zwavejs2mqtt/blob/4a6a5c5f1274763fd3aced4cae2c72ea060716b5/docs/guide/migrating.md).
        # To properly manage updates on writable values, propagate an event for both.
        if event_type == ZwaveValueChangedEvent and kwargs.get('value', {}).get('property_id') == 'currentValue':
            value = kwargs['value'].copy()
            target_value_id = f'{kwargs["node"]["node_id"]}-{value["command_class"]}-{value.get("endpoint", 0)}' \
                              f'-targetValue'
            kwargs['value'] = kwargs['node'].get('values', {}).get(target_value_id)

            if kwargs['value']:
                kwargs['value']['data'] = value['data']
                kwargs['node']['values'][target_value_id] = kwargs['value']
                evt = event_type(**kwargs)
                self._events_queue.put(evt)

    def on_mqtt_message(self):
        def handler(_, __, msg):
            if not msg.topic.startswith(self.events_topic):
                return

            topic = msg.topic[len(self.events_topic) + 1:].split('/').pop()
            data = msg.payload.decode()
            if not data:
                return

            try:
                data = json.loads(data)['data']
            except (ValueError, TypeError):
                pass

            try:
                if topic == 'node_value_updated':
                    self._dispatch_event(ZwaveValueChangedEvent, node=data[0], value=data[1])
                elif topic == 'node_metadata_updated':
                    self._dispatch_event(ZwaveNodeEvent, node=data[0])
                elif topic == 'node_sleep':
                    self._dispatch_event(ZwaveNodeAsleepEvent, node=data[0])
                elif topic == 'node_wakeup':
                    self._dispatch_event(ZwaveNodeAwakeEvent, node=data[0])
                elif topic == 'node_ready':
                    self._dispatch_event(ZwaveNodeReadyEvent, node=data[0])
                elif topic == 'node_removed':
                    self._dispatch_event(ZwaveNodeRemovedEvent, node=data[0])
            except Exception as e:
                self.logger.exception(e)

        return handler

    def run(self):
        super().run()
        self.logger.debug('Refreshing Z-Wave nodes')
        # noinspection PyUnresolvedReferences
        self._nodes = self.plugin.get_nodes().output

        while not self.should_stop():
            try:
                evt = self._events_queue.get(block=True, timeout=1)
            except Empty:
                continue

            self.bus.post(evt)


# vim:sw=4:ts=4:et:



import json
from typing import Optional, Iterable

import multiprocessing
import requests
from websocket import WebSocketApp

from platypush.context import get_bus
from platypush.message.event.chat.slack import SlackMessageReceivedEvent, SlackMessageDeletedEvent, \
    SlackMessageEditedEvent, SlackAppMentionReceivedEvent
from platypush.plugins import RunnablePlugin, action
from platypush.plugins.chat import ChatPlugin


class SlackPlugin(ChatPlugin, RunnablePlugin):
    """
    Plugin used to interact with Slack instances.

    You'll have to generate your own Slack app and tokens in order to use this plugin. Steps:

        - Create a new Slack app `here <https://api.slack.com/apps>` and associate a Slack workspace to it.
        - In the configuration panel of your app scroll to Socket Mode and select Enable Socket Mode.
        - Scroll to Event Subscriptions and select Enable Events.
        - Choose the type of events that you want to subscribe to. You can select bot events (i.e. when somebody in
          the channel mentions the name of your app) or any of the workspace events (e.g. creation of messages, user
          events etc.).
        - Scroll to App-Level Tokens and generate a new token with ``connections:write`` scope. This token will be
          used to receive Slack events over websocket.
        - Scroll to OAuth & Permissions and select the scopes that you want to enable. You may usually want to enable
          Bot Token Scopes -> ``app_mentions:read``, so the script can react when somebody mentions its name. You may
          also want to select the user scopes relevant to your application - e.g. read/write messages, manage users etc.
        - If you changed scopes and settings, you may have to reinstall the app in the workspace through the
          Install App menu.
        - Navigate to the Install App menu. If the app has been correctly installed in your workspace then you should
          see a Bot User OAuth Token, used to authenticate API calls performed as the app/bot. If you also granted
          user permissions to the app then you should also see a User OAuth Token on the page.

    Triggers:

        - :class:`platypush.message.event.chat.slack.SlackMessageReceivedEvent` when a message is received on a
            monitored channel.
        - :class:`platypush.message.event.chat.slack.SlackMessageEditedEvent` when a message is edited on a
            monitored channel.
        - :class:`platypush.message.event.chat.slack.SlackMessageDeletedEvent` when a message is deleted from a
            monitored channel.
        - :class:`platypush.message.event.chat.slack.SlackAppMentionReceivedEvent` when a message that mentions
            the app is received on a monitored channel.

    """

    _api_base_url = 'https://slack.com/api'

    def __init__(self, app_token: str, bot_token: str, user_token: Optional[str] = None, **kwargs):
        """
        :param app_token: Your Slack app token.
        :param bot_token: Bot OAuth token reported on the *Install App* menu.
        :param user_token: User OAuth token reported on the *Install App* menu.
        """
        super().__init__(**kwargs)
        self._app_token = app_token
        self._bot_token = bot_token
        self._user_token = user_token
        self._ws_url: Optional[str] = None
        self._ws_app: Optional[WebSocketApp] = None
        self._ws_listener: Optional[multiprocessing.Process] = None
        self._connected_event = multiprocessing.Event()
        self._disconnected_event = multiprocessing.Event()
        self._state_lock = multiprocessing.RLock()

    @classmethod
    def _url_for(cls, method: str) -> str:
        return f'{cls._api_base_url}/{method}'

    @action
    def send_message(self, channel: str, as_user: bool = False, text: Optional[str] = None,
                     blocks: Optional[Iterable[str]] = None, **kwargs):
        """
        Send a message to a channel.
        It requires a token with ``chat:write`` bot/user scope.

        :param channel: Channel ID or name.
        :param as_user: If true then the message will be sent as the authorized user, otherwise as the application bot
            (default: false).
        :param text: Text to be sent.
        :param blocks: Extra blocks to be added to the message (e.g. images, links, markdown). See
            `Slack documentation for blocks <https://api.slack.com/reference/block-kit/blocks>`_.
        """
        rs = requests.post(
            self._url_for('chat.postMessage'),
            headers={
                'Authorization': f'Bearer {self._user_token if as_user else self._bot_token}'
            },
            json={
                'channel': channel,
                'text': text,
                'blocks': blocks or [],
            }
        )

        try:
            rs.raise_for_status()
            rs = rs.json()
            assert rs.get('ok'), rs.get('error', rs.get('warning'))
        except Exception as e:
            raise AssertionError(e)

    def main(self):
        self._connect()
        stop_events = []

        while not any(stop_events):
            stop_events = self._should_stop.wait(timeout=1), self._disconnected_event.wait(timeout=1)

    def stop(self):
        if self._ws_app:
            self._ws_app.close()
            self._ws_app = None

        if self._ws_listener and self._ws_listener.is_alive():
            self.logger.info('Terminating websocket process')
            self._ws_listener.terminate()
            self._ws_listener.join(5)

        if self._ws_listener and self._ws_listener.is_alive():
            self.logger.warning('Terminating the websocket process failed, killing the process')
            self._ws_listener.kill()

        if self._ws_listener:
            self._ws_listener.join()
            self._ws_listener = None

        super().stop()

    def _connect(self):
        with self._state_lock:
            if self.should_stop() or self._connected_event.is_set():
                return

            self._ws_url = None
            rs = requests.post('https://slack.com/api/apps.connections.open', headers={
                'Authorization': f'Bearer {self._app_token}',
            })

            try:
                rs.raise_for_status()
            except:   # lgtm [py/catch-base-exception]
                if rs.status_code == 401 or rs.status_code == 403:
                    self.logger.error('Unauthorized/Forbidden Slack API request, stopping the service')
                    self.stop()
                    return

                raise

            rs = rs.json()
            assert rs.get('ok')
            self._ws_url = rs.get('url')
            self._ws_app = WebSocketApp(self._ws_url,
                                        on_open=self._on_open(),
                                        on_message=self._on_msg(),
                                        on_error=self._on_error(),
                                        on_close=self._on_close())

            def server():
                self._ws_app.run_forever()

            self._ws_listener = multiprocessing.Process(target=server)
            self._ws_listener.start()

    def _on_open(self):
        def hndl(*_):
            with self._state_lock:
                self._disconnected_event.clear()
                self._connected_event.set()
            self.logger.info('Connected to the Slack websocket')

        return hndl

    @staticmethod
    def _send_ack(ws: WebSocketApp, msg):
        envelope_id = msg.get('envelope_id')
        if envelope_id:
            # Send ACK
            ws.send(json.dumps({
                'envelope_id': envelope_id,
            }))

    def _on_msg(self):
        def hndl(*args):
            ws = args[0] if len(args) > 1 else self._ws_app
            data = json.loads(args[1] if len(args) > 1 else args[0])
            output_event = None
            self._send_ack(ws, data)

            if data['type'] == 'events_api':
                event = data.get('payload', {}).get('event', {})
                event_args = {}

                if event['type'] == 'app_mention':
                    output_event = SlackAppMentionReceivedEvent(
                        text=event['text'],
                        user=event['user'],
                        channel=event['channel'],
                        team=event['team'],
                        timestamp=event['event_ts'],
                        icons=event.get('icons'),
                        blocks=event.get('blocks')
                    )
                elif event['type'] == 'message':
                    msg = event.copy()
                    prev_msg = event.get('previous_message')
                    event_type = SlackMessageReceivedEvent

                    if event.get('subtype') == 'message_deleted':
                        msg = prev_msg
                        event_type = SlackMessageDeletedEvent
                        event_args['timestamp'] = event['deleted_ts']
                    else:
                        event_args['timestamp'] = msg.get('ts')
                        if event.get('subtype') == 'message_changed':
                            msg = msg.get('message', msg)
                            event_args['previous_message'] = prev_msg
                            event_type = SlackMessageEditedEvent

                    event_args.update({
                        'text': msg.get('text'),
                        'user': msg.get('user'),
                        'channel': msg.get('channel', event.get('channel')),
                        'team': msg.get('team'),
                        'icons': msg.get('icons'),
                        'blocks': msg.get('blocks'),
                    })

                    output_event = event_type(**event_args)

            if output_event:
                get_bus().post(output_event)

        return hndl

    def _on_error(self):
        def hndl(*args):
            error = args[1] if len(args) > 1 else args[0]
            ws = args[0] if len(args) > 1 else None
            self.logger.warning('Slack websocket error: {}'.format(error))
            if ws:
                ws.close()

        return hndl

    def _on_close(self):
        def hndl(*_):
            with self._state_lock:
                self._disconnected_event.set()
                self._connected_event.clear()
            self.logger.warning('Slack websocket connection closed')

        return hndl



from marshmallow import fields
from marshmallow.schema import Schema

from platypush.schemas import DateTime


class RssFeedEntrySchema(Schema):
    feed_title = fields.String(metadata=dict(description='Feed title'))
    feed_url = fields.URL(
        required=True,
        metadata=dict(
            description='URL of the feed',
            example='https://some-website/rss',
        )
    )

    id = fields.String(
        required=True,
        metadata=dict(
            description='Feed entry ID',
            example='1234',
        )
    )

    url = fields.URL(
        required=True,
        metadata=dict(
            description='URL of the feed entry',
            example='https://some-website/articles/1234',
        )
    )

    published = DateTime(
        required=True,
        metadata=dict(description='Entry published time')
    )

    title = fields.String(metadata=dict(description='Feed entry title'))
    summary = fields.String(metadata=dict(description='Feed entry summary'))
    content = fields.String(metadata=dict(description='Feed entry content'))



#!/usr/bin/env python

import os
from setuptools import setup, find_packages


def path(fname=''):
    return os.path.abspath(os.path.join(os.path.dirname(__file__), fname))


def readfile(fname):
    with open(path(fname)) as f:
        return f.read()


# noinspection PyShadowingBuiltins
def pkg_files(dir):
    paths = []
    # noinspection PyShadowingNames
    for (path, _, files) in os.walk(dir):
        for file in files:
            paths.append(os.path.join('..', path, file))
    return paths


plugins = pkg_files('platypush/plugins')
backend = pkg_files('platypush/backend')

setup(
    name="platypush",
    version="0.24.0",
    author="Fabio Manganiello",
    author_email="info@fabiomanganiello.com",
    description="Platypush service",
    license="MIT",
    python_requires='>= 3.6',
    keywords="home-automation automation iot mqtt websockets redis dashboard notifications",
    url="https://platypush.tech",
    packages=find_packages(),
    include_package_data=True,
    entry_points={
        'console_scripts': [
            'platypush=platypush:main',
            'platydock=platypush.platydock:main',
        ],
    },
    scripts=['bin/platyvenv'],
    long_description=readfile('README.md'),
    long_description_content_type='text/markdown',
    classifiers=[
        "Topic :: Utilities",
        "License :: OSI Approved :: MIT License",
        "Development Status :: 4 - Beta",
    ],
    install_requires=[
        'pyyaml',
        'redis',
        'requests',
        'croniter',
        'sqlalchemy',
        'websockets',
        'websocket-client',
        'wheel',
        'zeroconf>=0.27.0',
        'tz',
        'python-dateutil',
        'rsa',
        'marshmallow',
        'frozendict',
        'flask',
        'bcrypt',
        'python-magic',
    ],
    extras_require={
        # Support for thread custom name
        'threadname': ['python-prctl'],
        # Support for Kafka backend and plugin
        'kafka': ['kafka-python'],
        # Support for Pushbullet backend and plugin
        'pushbullet': [
            'pushbullet.py @ https://github.com/rbrcsk/pushbullet.py/tarball/master'
        ],
        # Support for HTTP backend over uWSGI
        'http': ['gunicorn'],
        # Support for MQTT backends
        'mqtt': ['paho-mqtt'],
        # Support for RSS feeds parser
        'rss': ['feedparser', 'defusedxml'],
        # Support for PDF generation
        'pdf': ['weasyprint'],
        # Support for Philips Hue plugin
        'hue': ['phue'],
        # Support for MPD/Mopidy music server plugin and backend
        'mpd': ['python-mpd2'],
        # Support for Google text2speech plugin
        'google-tts': [
            'oauth2client',
            'google-api-python-client',
            'google-cloud-texttospeech',
        ],
        # Support for OMXPlayer plugin
        'omxplayer': ['omxplayer-wrapper'],
        # Support for YouTube
        'youtube': ['youtube-dl'],
        # Support for torrents download
        'torrent': ['python-libtorrent-bin'],
        # Generic support for cameras
        'camera': ['numpy', 'Pillow'],
        # Support for RaspberryPi camera
        'picamera': ['picamera', 'numpy', 'Pillow'],
        # Support for inotify file monitors
        'inotify': ['inotify'],
        # Support for Google Assistant
        'google-assistant-legacy': ['google-assistant-library'],
        'google-assistant': ['google-assistant-sdk[samples]'],
        # Support for the Google APIs
        'google': ['oauth2client', 'google-api-python-client'],
        # Support for Last.FM scrobbler plugin
        'lastfm': ['pylast'],
        # Support for custom hotword detection
        'hotword': ['snowboy'],
        'snowboy': ['snowboy'],
        # Support for real-time MIDI events
        'midi': ['rtmidi'],
        # Support for RaspberryPi GPIO
        'rpi-gpio': ['RPi.GPIO'],
        # Support for MCP3008 analog-to-digital converter plugin
        'mcp3008': ['adafruit-mcp3008'],
        # Support for smart cards detection
        'scard': ['pyscard'],
        # Support for serial port plugin
        'serial': ['pyserial'],
        # Support for ICal calendars
        'ical': ['icalendar'],
        # Support for joystick backend
        'joystick': ['inputs'],
        # Support for Kodi plugin
        'kodi': ['kodi-json'],
        # Support for Plex plugin
        'plex': ['plexapi'],
        # Support for Chromecast plugin
        'chromecast': ['pychromecast'],
        # Support for sound devices
        'sound': ['sounddevice', 'soundfile', 'numpy'],
        # Support for web media subtitles
        'subtitles': [
            'webvtt-py',
            'python-opensubtitles @ https://github.com/agonzalezro/python-opensubtitles/tarball/master',
        ],
        # Support for mpv player plugin
        'mpv': ['python-mpv'],
        # Support for NFC tags
        'nfc': ['nfcpy>=1.0', 'ndeflib'],
        # Support for enviropHAT
        'envirophat': ['envirophat'],
        # Support for GPS
        'gps': ['gps'],
        # Support for BME280 environment sensor
        'bme280': ['pimoroni-bme280'],
        # Support for LTR559 light/proximity sensor
        'ltr559': ['ltr559'],
        # Support for VL53L1X laser ranger/distance sensor
        'vl53l1x': ['smbus2', 'vl53l1x'],
        # Support for Dropbox integration
        'dropbox': ['dropbox'],
        # Support for Leap Motion backend
        'leap': [
            'leap-sdk @ https://github.com/BlackLight/leap-sdk-python3/tarball/master'
        ],
        # Support for Flic buttons
        'flic': [
            'flic @ https://github.com/50ButtonsEach/fliclib-linux-hci/tarball/master'
        ],
        # Support for Alexa/Echo plugin
        'alexa': ['avs @ https://github.com/BlackLight/avs/tarball/master'],
        # Support for bluetooth devices
        'bluetooth': [
            'pybluez',
            'gattlib',
            'pyobex @ https://github.com/BlackLight/PyOBEX/tarball/master',
        ],
        # Support for TP-Link devices
        'tplink': ['pyHS100'],
        # Support for PMW3901 2-Dimensional Optical Flow Sensor
        'pmw3901': ['pmw3901'],
        # Support for MLX90640 thermal camera
        'mlx90640': ['Pillow'],
        # Support for machine learning models and cameras over OpenCV
        'cv': ['opencv-python', 'numpy', 'Pillow'],
        # Support for the generation of HTML documentation from docstring
        'htmldoc': ['docutils'],
        # Support for Node-RED integration
        'nodered': ['pynodered'],
        # Support for Todoist integration
        'todoist': ['todoist-python'],
        # Support for Trello integration
        'trello': ['py-trello'],
        # Support for Google Pub/Sub
        'google-pubsub': ['google-cloud-pubsub'],
        # Support for Google Translate
        'google-translate': ['google-cloud-translate'],
        # Support for keyboard/mouse plugin
        'inputs': ['pyuserinput'],
        # Support for Buienradar weather forecast
        'buienradar': ['buienradar'],
        # Support for Telegram integration
        'telegram': ['python-telegram-bot'],
        # Support for Arduino integration
        'arduino': ['pyserial', 'pyfirmata2'],
        # Support for CUPS printers management
        'cups': ['pycups'],
        # Support for Graphite integration
        'graphite': ['graphyte'],
        # Support for CPU and memory monitoring and info
        'sys': ['py-cpuinfo', 'psutil'],
        # Support for nmap integration
        'nmap': ['python-nmap'],
        # Support for zigbee2mqtt
        'zigbee': ['paho-mqtt'],
        # Support for Z-Wave
        'zwave': ['python-openzwave'],
        # Support for Mozilla DeepSpeech speech-to-text engine
        'deepspeech': ['deepspeech', 'numpy', 'sounddevice'],
        # Support for PicoVoice hotword detection engine
        'picovoice-hotword': ['pvporcupine'],
        # Support for PicoVoice speech-to-text engine
        'picovoice-speech': ['pvcheetah @ git+https://github.com/BlackLight/cheetah'],
        # Support for OTP (One-Time Password) generation
        'otp': ['pyotp'],
        # Support for Linode integration
        'linode': ['linode_api4'],
        # Support for QR codes
        'qrcode': ['numpy', 'qrcode[pil]', 'Pillow', 'pyzbar'],
        # Support for Tensorflow
        'tensorflow': ['numpy', 'tensorflow>=2.0', 'keras', 'pandas'],
        # Support for Samsung TizenOS-based smart TVs
        'samsungtv': ['samsungtvws'],
        # Support for SSH integration
        'ssh': ['paramiko'],
        # Support for clipboard integration
        'clipboard': ['pyperclip'],
        # Support for luma.oled display drivers
        'luma-oled': ['luma.oled @ git+https://github.com/rm-hull/luma.oled'],
        # Support for DBus integration
        'dbus': ['pydbus', 'defusedxml'],
        # Support for Twilio integration
        'twilio': ['twilio'],
        # Support for DHT11/DHT22/AM2302 temperature/humidity sensors
        'dht': [
            'Adafruit_Python_DHT @ git+https://github.com/adafruit/Adafruit_Python_DHT'
        ],
        # Support for LCD display integration
        'lcd': ['RPi.GPIO', 'RPLCD'],
        # Support for IMAP mail integration
        'imap': ['imapclient'],
        # Support for NextCloud integration
        'nextcloud': ['nextcloud-api-wrapper'],
        # Support for VLC integration
        'vlc': ['python-vlc'],
        # Support for SmartThings integration
        'smartthings': ['pysmartthings', 'aiohttp'],
        # Support for file.monitor backend
        'filemonitor': ['watchdog'],
        # Support for Adafruit PCA9685 PWM controller
        'pca9685': ['adafruit-python-shell', 'adafruit-circuitpython-pca9685'],
        # Support for ngrok integration
        'ngrok': ['pyngrok'],
        # Support for IRC integration
        'irc': ['irc'],
        # Support for the Matrix integration
        'matrix': ['matrix-nio'],
    },
)
